import logging

from abc import ABCMeta, abstractmethod

from gordo_dataset.file_system import FileSystem, FileInfo
from .file_type import FileType, ParquetFileType, CsvFileType, TimeSeriesColumns
from .partition import Partition, YearPartition, MonthPartition

from typing import Iterable, Optional, List, Tuple, Type, Dict, cast

from ..exceptions import ConfigException


logger = logging.getLogger(__name__)

time_series_columns = TimeSeriesColumns("Time", "Value", "Status")


class NcsFileType(metaclass=ABCMeta):
    """
    Represents logic about finding files of one particular type for ``NcsLookup``
    """

    @property
    @abstractmethod
    def file_type(self) -> FileType:
        ...

    @property
    @abstractmethod
    def partition_type(self) -> Type[Partition]:
        ...

    def check_partition(self, partition: Partition):
        return isinstance(partition, self.partition_type)

    @abstractmethod
    def paths(
        self,
        fs: FileSystem,
        tag_dir: str,
        tag_name: str,
        partitions: Iterable[Partition],
    ) -> Iterable[Tuple[Partition, str]]:
        """
        Possible file paths for this file type. These paths should be relational to the tag directory

        Parameters
        ----------
        fs: FileSystem
        tag_dir: str
        tag_name: str
        partitions: Iterable[Partition]

        Returns
        -------
        Iterable[Tuple[Partition, str]]

        """
        ...


class NcsMonthlyParquetFileType(NcsFileType):
    """
    NCS monthly parquet files finder
    """

    def __init__(self):
        self._file_type = ParquetFileType(time_series_columns)
        self._partition_type = YearPartition

    @property
    def file_type(self) -> FileType:
        return self._file_type

    @property
    def partition_type(self) -> Type[Partition]:
        return MonthPartition

    def paths(
        self,
        fs: FileSystem,
        tag_dir: str,
        tag_name: str,
        partitions: Iterable[Partition],
    ) -> Iterable[Tuple[Partition, str]]:
        file_extension = self._file_type.file_extension
        month_partitions: List[MonthPartition] = []
        for partition in partitions:
            if not self.check_partition(partition):
                raise NotImplementedError()
            month_partitions.append(cast(MonthPartition, partition))
        tag_name_upper = tag_name.upper()
        dirs_by_years: Dict[int, List[Tuple[str, Optional[FileInfo]]]] = {}
        for partition in month_partitions:
            if partition.year not in dirs_by_years:
                dir_path = fs.join(tag_dir, "parquet", str(partition.year))
                try:
                    dirs_by_years[partition.year] = list(fs.ls(dir_path))
                except FileNotFoundError:
                    logger.info("File not found '%s'", dir_path)
                    break
            files_list = dirs_by_years[partition.year]
            for path, file_info in files_list:
                _, file_name = fs.split(path)
                if cast(FileInfo, file_info).isfile():
                    file_name_expected = f"{tag_name_upper}_{partition.year}{partition.month:02d}{file_extension.upper()}"
                    if file_name.upper() == file_name_expected:
                        yield partition, path


class NcsYearlyParquetFileType(NcsFileType):
    """
    NCS yearly parquet files finder
    """

    def __init__(self):
        self._file_type = ParquetFileType(time_series_columns)
        self._partition_type = YearPartition

    @property
    def file_type(self) -> FileType:
        return self._file_type

    @property
    def partition_type(self) -> Type[Partition]:
        return YearPartition

    def paths(
        self,
        fs: FileSystem,
        tag_dir: str,
        tag_name: str,
        partitions: Iterable[Partition],
    ) -> Iterable[Tuple[Partition, str]]:
        file_extension = self._file_type.file_extension
        year_partitions: List[YearPartition] = []
        for partition in partitions:
            if not self.check_partition(partition):
                raise NotImplementedError()
            year_partitions.append(cast(YearPartition, partition))
        tag_name_upper = tag_name.upper()
        for partition in year_partitions:
            dir_path = fs.join(tag_dir, "parquet")
            try:
                files_list = list(fs.ls(dir_path))
            except FileNotFoundError:
                logger.info("File not found '%s'", dir_path)
                break
            for path, file_info in files_list:
                _, file_name = fs.split(path)
                if cast(FileInfo, file_info).isfile():
                    file_name_expected = (
                        f"{tag_name_upper}_{partition.year}{file_extension.upper()}"
                    )
                    if file_name.upper() == file_name_expected:
                        yield partition, path


class NcsCsvFileType(NcsFileType):
    """
    NCS CSV files finder
    """

    def __init__(self):
        header = ["Sensor", "Value", "Time", "Status"]
        self._file_type = CsvFileType(header, time_series_columns)
        self._partition_type = YearPartition

    @property
    def file_type(self) -> FileType:
        return self._file_type

    @property
    def partition_type(self) -> Type[Partition]:
        return self._partition_type

    def paths(
        self,
        fs: FileSystem,
        tag_dir: str,
        tag_name: str,
        partitions: Iterable[Partition],
    ) -> Iterable[Tuple[Partition, str]]:
        # TODO Make this case insensitive for tag_name
        file_extension = self._file_type.file_extension
        for partition in partitions:
            if not self.check_partition(partition):
                raise NotImplementedError()
            path = f"{tag_name}_{partition.year}{file_extension}"
            yield partition, fs.join(tag_dir, path)


ncs_file_types: Dict[str, Type[NcsFileType]] = {
    "parquet": NcsMonthlyParquetFileType,
    "yearly_parquet": NcsYearlyParquetFileType,
    "csv": NcsCsvFileType,
}

DEFAULT_TYPE_NAMES: List[str] = ["parquet", "yearly_parquet", "csv"]


def load_ncs_file_types(
    type_names: Optional[Iterable[str]] = None,
) -> List[NcsFileType]:
    """
    Returns list of ``NcsFileType`` instances from names of those types

    Parameters
    ----------
    type_names: Optional[Iterable[str]]
        List of ``NcsFileType`` names. Only supporting `parquet` and `csv` names values

    Returns
    -------
    List[NcsFileType]

    """
    if type_names is None:
        type_names = DEFAULT_TYPE_NAMES
    result = []
    for type_name in type_names:
        if type_name not in ncs_file_types:
            raise ConfigException("Can not find file type '%s'" % type_name)
        result.append(ncs_file_types[type_name]())
    return result



import logging

from urllib.parse import quote
from dataclasses import dataclass

from gordo_dataset.file_system import FileSystem
from gordo_dataset.sensor_tag import SensorTag, extract_tag_name
from gordo_dataset.exceptions import ConfigException
from gordo_dataset.assets_config import AssetsConfig, PathSpec
from .file_type import FileType
from .ncs_file_type import NcsFileType, load_ncs_file_types
from .constants import DEFAULT_MAX_FILE_SIZE
from .partition import Partition, YearPartition

from typing import (
    List,
    Iterable,
    Tuple,
    Optional,
    Dict,
    Iterator,
    Union,
    Set,
    TypeVar,
    cast,
)
from collections import defaultdict, OrderedDict
from concurrent.futures import ThreadPoolExecutor

logger = logging.getLogger(__name__)

T = TypeVar("T")


@dataclass(frozen=True)
class Location:
    """
    Represents location of the tag in the data lake
    """

    path: str
    file_type: FileType
    partition: Optional[Partition] = None


@dataclass(frozen=True)
class TagLocations:
    """
    Locations of the tags for each partition
    """

    orig_tag: Union[str, SensorTag]
    tag: Optional[SensorTag] = None
    locations: Optional[Dict[Partition, Location]] = None

    @property
    def tag_name(self):
        return extract_tag_name(self.orig_tag)

    def available(self) -> bool:
        return self.tag is not None and self.locations is not None

    def partitions(self) -> List[Partition]:
        if not self.available():
            return []
        return sorted(cast(Dict[Partition, Location], self.locations).keys())

    def get_location(self, partition: Union[int, Partition]) -> Optional[Location]:
        curr_partition: Partition = cast(Partition, partition)
        if type(partition) is int:
            curr_partition = YearPartition(cast(int, partition))
        if not self.available():
            return None
        return cast(Dict[Partition, Location], self.locations).get(curr_partition)

    def __iter__(self) -> Iterator[Tuple[SensorTag, Partition, Location]]:
        if self.available():
            locations = cast(Dict[Partition, Location], self.locations)
            for partition in self.partitions():
                yield cast(SensorTag, self.tag), cast(Partition, partition), locations[
                    partition
                ]


PathSpecWithTag = Tuple[PathSpec, Union[str, SensorTag]]


TAG_DIRS_OPTIONAL_SUFFIXES = "."


class NcsLookup:
    """
    Class which could be used for finding tags data in the data lake storage
    """

    @classmethod
    def create(
        cls,
        storage: FileSystem,
        ncs_type_names: Optional[Iterable[str]] = None,
        storage_name: Optional[str] = None,
        max_file_size: Optional[int] = DEFAULT_MAX_FILE_SIZE,
    ) -> "NcsLookup":
        ncs_file_types = load_ncs_file_types(ncs_type_names)
        return cls(storage, ncs_file_types, storage_name, max_file_size=max_file_size)

    def __init__(
        self,
        storage: FileSystem,
        ncs_file_types: List[NcsFileType],
        storage_name: Optional[str] = None,
        max_file_size: Optional[int] = DEFAULT_MAX_FILE_SIZE,
    ):
        self.storage = storage
        self.ncs_file_types = ncs_file_types
        if storage_name is None:
            storage_name = storage.name
        self.storage_name = storage_name
        self.max_file_size = max_file_size

    @staticmethod
    def quote_tag_name(tag_name: str) -> str:
        return quote(tag_name, safe=" ")

    @staticmethod
    def _validate_tag_equality(
        tag1: Union[str, SensorTag], tag2: Union[str, SensorTag]
    ):
        # TODO better docstring
        type_tag1, type_tag2 = type(tag1), type(tag2)
        if type_tag1 is SensorTag and type_tag2 is SensorTag:
            if cast(SensorTag, tag1).asset != cast(SensorTag, tag2).asset:
                raise ValueError(
                    "Tags %s and %s with the same name but different assets"
                    % (repr(tag1), repr(tag2))
                )
        for _type in (str, SensorTag):
            if type_tag1 is _type:
                if type_tag2 is not _type:
                    tag_name1 = extract_tag_name(tag1)
                    tag_name2 = extract_tag_name(tag2)
                    if tag_name1 == tag_name2:
                        raise ValueError(
                            "Tags %s and %s has different type but the same name"
                            % (repr(tag1), repr(tag2))
                        )

    def tag_dirs_lookup(
        self, specs: List[PathSpecWithTag]
    ) -> Iterable[Tuple[str, Optional[Tuple[str, SensorTag]]]]:
        # TODO better docstring
        # TODO split this method
        """
        Takes list of tags and find directories related to them in the data lake storage

        Parameters
        ----------
        specs: List[Tuple[PathSpec, Union[str, SensorTag]]]

        Returns
        -------

        """
        storage = self.storage
        path_tag_and_specs: Dict[str, Dict[str, PathSpecWithTag]] = {}
        tag_paths: Dict[str, Set[str]] = defaultdict(set)
        unique_tags: Dict[str, Union[str, SensorTag]] = {}
        for path_spec, tag in specs:
            base_dir = path_spec.full_path(storage)
            if base_dir not in path_tag_and_specs:
                path_tag_and_specs[base_dir] = {}
            tag_and_spec = path_tag_and_specs[base_dir]
            tag_name_orig: str = extract_tag_name(tag)
            if tag_name_orig in unique_tags:
                self._validate_tag_equality(tag, unique_tags[tag_name_orig])
            else:
                unique_tags[tag_name_orig] = tag
            tag_name = self.quote_tag_name(tag_name_orig).upper()
            if tag_name in tag_and_spec:
                raise ValueError(
                    "Found duplicate for tag '%s' in '%s' with spec '%s'"
                    % (tag_name_orig, base_dir, tag_and_spec[tag_name])
                )
            tag_and_spec[tag_name] = (path_spec, tag)
            tag_paths[tag_name].add(base_dir)

        found_tags: Dict[str, Dict[str, Optional[Tuple[str, PathSpecWithTag]]]] = {
            tag_name: OrderedDict() for tag_name in tag_paths.keys()
        }

        def check_if_tag_already_found(
            check_tag_name: str,
        ) -> Optional[Tuple[str, Optional[Tuple[str, SensorTag]]]]:
            if check_tag_name not in tag_paths:
                return None
            if len(found_tags[tag_name]) == len(tag_paths[tag_name]):
                _, orig_tag = tag_and_specs[tag_name]
                orig_tag_name = extract_tag_name(orig_tag)
                for path_spec_with_tag in found_tags[tag_name].values():
                    if path_spec_with_tag is not None:
                        found_path, (found_path_spec, found_tag) = path_spec_with_tag
                        sensor_tag = self._normalize_tag(
                            found_tag, found_path_spec.asset
                        )
                        return orig_tag_name, (found_path, sensor_tag)
                return orig_tag_name, None
            return None

        for base_dir, tag_and_specs in path_tag_and_specs.items():
            for path, file_info in self.storage.ls(base_dir):
                if file_info is not None and file_info.isdir():
                    dir_path, file_name = self.storage.split(path)
                    file_tag_name = file_name.upper()
                    found_tag_name: Optional[str] = None
                    if file_tag_name in tag_and_specs:
                        found_tag_name = file_tag_name
                    else:
                        for tag_suffix in TAG_DIRS_OPTIONAL_SUFFIXES:
                            if file_tag_name + tag_suffix in tag_and_specs:
                                found_tag_name = file_tag_name + tag_suffix
                                break
                    if found_tag_name is not None:
                        if base_dir in found_tags[found_tag_name]:
                            logger.error(
                                "Found duplicate tag dir '%s' in '%s'",
                                found_tag_name,
                                base_dir,
                            )
                            continue
                        found_tags[found_tag_name][base_dir] = (
                            path,
                            tag_and_specs[found_tag_name],
                        )
                        result = check_if_tag_already_found(found_tag_name)
                        if result is not None:
                            yield result
                            del tag_paths[found_tag_name]
            for tag_name, tag_path_specs in tag_and_specs.items():
                if base_dir not in found_tags[tag_name]:
                    found_tags[tag_name][base_dir] = None
                result = check_if_tag_already_found(tag_name)
                if result is not None:
                    yield result
                    del tag_paths[tag_name]

    def _validate_file(self, full_path: str):
        storage = self.storage
        if self.max_file_size is not None:
            file_info = storage.info(full_path)
            if file_info.size > self.max_file_size:
                logger.debug(
                    "Size of file '%s' is %d bytes that bigger than the maximum file size %d bytes"
                    % (full_path, file_info.size, self.max_file_size)
                )
                return False
        return True

    def files_lookup(
        self,
        tag_dir: str,
        tag: SensorTag,
        partitions: Iterable[Partition],
        orig_tag: Optional[Union[str, SensorTag]] = None,
    ) -> TagLocations:
        """
        Finds files (possible parquet or CSV) in tag directory in the data lake storage

        Parameters
        ----------
        tag_dir: str
        tag: SensorTag
        partitions: Iterable[Partition]
            List of partitions for finding files
        orig_tag: Union[str, SensorTag]

        Returns
        -------
        TagLocations

        """
        storage = self.storage
        ncs_file_types = self.ncs_file_types
        tag_name = self.quote_tag_name(tag.name)
        locations = {}
        for partition in partitions:
            found = False
            for ncs_file_type in ncs_file_types:
                if ncs_file_type.check_partition(partition):
                    for path_partition, full_path in ncs_file_type.paths(
                        storage, tag_dir, tag_name, [partition]
                    ):
                        if storage.exists(full_path) and self._validate_file(full_path):
                            file_type = ncs_file_type.file_type
                            locations[partition] = Location(
                                full_path, file_type, path_partition
                            )
                            found = True
                            break
                    if found:
                        break
        orig_tag = tag if orig_tag is None else orig_tag
        return TagLocations(
            orig_tag=orig_tag, tag=tag, locations=locations if locations else None
        )

    @staticmethod
    def _normalize_tag(tag: Union[str, SensorTag], asset: Optional[str] = None):
        if type(tag) is str:
            if asset is None:
                raise ConfigException("asset should be specified along with base_name")
            return SensorTag(tag, asset)
        return tag

    def assets_config_tags_lookup(
        self,
        asset_config: AssetsConfig,
        tags: List[Union[str, SensorTag]],
        base_dir: Optional[str] = None,
        asset: Optional[str] = None,
    ) -> Iterable[Tuple[str, Optional[Tuple[str, SensorTag]]]]:
        # TODO better docstring
        """
        Takes assets paths from ``AssetsConfig`` and find tag paths in the data lake storage

        Parameters
        ----------
        asset_config: AssetsConfig
        tags: List[Union[str, SensorTag]]
        base_dir: Optional[str]
        asset: Optional[str]

        Returns
        -------

        """
        storage = self.storage
        storage_name = storage.name
        specs: List[PathSpecWithTag] = []
        if not base_dir:
            for tag in tags:
                tag_name: str
                if type(tag) is SensorTag:
                    sensor_tag = cast(SensorTag, tag)
                    tag_name = sensor_tag.name
                    asset = sensor_tag.asset
                else:
                    tag_name = cast(str, tag)
                path_specs = asset_config.get_paths(storage_name, tag_name, asset)
                if not path_specs:
                    raise ValueError(
                        "Cant find tag '%s' in storage '%s'", tag_name, storage_name
                    )
                for path_spec in path_specs:
                    specs.append((path_spec, tag))
        else:
            path_spec = PathSpec(cast(str, asset), base_dir, "")
            for tag in tags:
                sensor_tag = self._normalize_tag(tag, asset)
                specs.append((path_spec, sensor_tag))
        yield from self.tag_dirs_lookup(specs)

    def _thread_pool_lookup_mapper(
        self,
        tag_lookup: Tuple[str, Optional[Tuple[str, SensorTag]]],
        partitions: List[Partition],
        tag_by_name: Dict[str, Union[str, SensorTag]],
    ) -> TagLocations:
        tag_name, tag_with_path = tag_lookup
        orig_tag = tag_by_name.get(tag_name, tag_name)
        if tag_with_path is not None:
            tag_dir, tag = tag_with_path
            return self.files_lookup(tag_dir, tag, partitions, orig_tag=orig_tag)
        else:
            return TagLocations(orig_tag)

    @staticmethod
    def _inf_iterator(v: T) -> Iterable[T]:
        while True:
            yield v

    @classmethod
    def _tags_by_names(
        cls, tags: List[Union[str, SensorTag]]
    ) -> Dict[str, Union[str, SensorTag]]:
        orig_tags: Dict[str, Union[str, SensorTag]] = {}
        for tag in tags:
            tag_name = extract_tag_name(tag)
            if tag_name in orig_tags:
                cls._validate_tag_equality(tag, orig_tags[tag_name])
            else:
                orig_tags[tag_name] = tag
        return orig_tags

    def lookup(
        self,
        asset_config: AssetsConfig,
        tags: List[Union[str, SensorTag]],
        partitions: Iterable[Partition],
        threads_count: int = 1,
        base_dir: Optional[str] = None,
        asset: Optional[str] = None,
    ) -> Iterable[TagLocations]:
        """
        Takes assets paths from ``AssetsConfig`` and find tags files paths in the data lake storage

        Parameters
        ----------
        asset_config: AssetsConfig
        tags: List[SensorTag]
        partitions: Iterable[Partition]
        threads_count: int
            Number of threads for internal `ThreadPool`. Do not uses thread pool if 1
        base_dir: Optional[str]
        asset: Optional[str]
            Should be specified when base_dir is not None

        Returns
        -------

        """
        if not threads_count or threads_count < 1:
            raise ConfigException("thread_count should bigger or equal to 1")
        multi_thread = threads_count > 1

        tag_by_names = self._tags_by_names(tags)
        tag_lookups = self.assets_config_tags_lookup(
            asset_config, tags, base_dir=base_dir, asset=asset
        )
        partitions_tuple = tuple(partitions)
        if multi_thread:
            with ThreadPoolExecutor(max_workers=threads_count) as executor:
                result = executor.map(
                    self._thread_pool_lookup_mapper,
                    tag_lookups,
                    self._inf_iterator(partitions_tuple),
                    self._inf_iterator(tag_by_names),
                )
                for tag_locations in result:
                    yield tag_locations
        else:
            for tag_name, path_with_tag in tag_lookups:
                orig_tag = tag_by_names.get(tag_name, tag_name)
                if path_with_tag is not None:
                    tag_dir, tag = path_with_tag
                    yield self.files_lookup(
                        tag_dir, tag, partitions_tuple, orig_tag=orig_tag
                    )
                else:
                    yield TagLocations(orig_tag)



import logging
from collections import namedtuple
from typing import Union, List, Dict, Optional, cast
from .assets_config import AssetsConfig


logger = logging.getLogger(__name__)

SensorTag = namedtuple("SensorTag", ["name", "asset"])

Tag = Union[str, SensorTag]


def load_sensor_tag(
    sensor: Union[Dict, List],
) -> SensorTag:
    # TODO better docstring
    sensor_tag: Union[str, SensorTag]
    if isinstance(sensor, Dict):
        for key in ("name", "asset"):
            if key not in sensor:
                raise SensorTagNormalizationError(
                    "Sensor representation %s does not have '%s'" % (repr(sensor), key)
                )
        return SensorTag(sensor["name"], sensor["asset"])

    elif isinstance(sensor, List):
        if len(sensor) != 2:
            raise SensorTagNormalizationError(
                "Sensor representation list %s does not have enough elements"
                % repr(sensor)
            )
        return SensorTag(sensor[0], sensor[1])

    else:
        raise SensorTagNormalizationError(
            f"Unable to load sensor_tag from {sensor} with type {type(sensor)}"
        )


def normalize_sensor_tag(
    assets_config: AssetsConfig,
    sensor: Union[Dict, List, str, SensorTag],
    asset: str = None,
) -> Union[str, SensorTag]:
    # TODO better docstring
    sensor_tag: Union[str, SensorTag]

    if isinstance(sensor, SensorTag):
        sensor_tag = sensor

    elif isinstance(sensor, Dict) or isinstance(sensor, List):
        sensor_tag = load_sensor_tag(sensor)

    elif isinstance(sensor, str):
        if asset is not None:
            sensor_tag = SensorTag(sensor, asset)
        else:
            sensor_tag = sensor

    else:
        raise SensorTagNormalizationError(
            f"Sensor {sensor} with type {type(sensor)} cannot be converted to a valid "
            f"SensorTag"
        )

    if isinstance(sensor_tag, SensorTag):
        asset_name = assets_config.prepare_asset_name(sensor_tag.asset)
        if sensor_tag.asset != asset_name:
            return SensorTag(sensor_tag.name, asset_name)
    return sensor_tag


def legacy_normalize_sensor_tag(
    assets_config: AssetsConfig,
    storage: str,
    sensor: Union[str, SensorTag],
    asset: Optional[str] = None,
) -> SensorTag:
    # TODO better docstring
    if isinstance(sensor, SensorTag):
        return sensor

    path_specs = assets_config.get_paths(storage, sensor, asset)
    if len(path_specs) == 1:
        path_spec = path_specs[0]
        return SensorTag(sensor, path_spec.asset)

    considered_assets = [path_spec.asset for path_spec in path_specs]
    raise SensorTagNormalizationError(
        "Unable to find exact asset for sensor %s in the assets config. Considered assets: %s"
        % (repr(sensor), ", ".join(repr(asset) for asset in considered_assets))
    )


def to_list_of_strings(sensor_tag_list: List[SensorTag]):
    return [sensor_tag.name for sensor_tag in sensor_tag_list]


def extract_tag_name(tag: Tag) -> str:
    if type(tag) is str:
        return cast(str, tag)
    else:
        return cast(SensorTag, tag).name


def extract_tag_asset(tag: Tag) -> Optional[str]:
    if type(tag) is not str:
        return cast(SensorTag, tag).asset
    return None


def tag_to_json(tag: Tag) -> Union[str, dict]:
    if type(tag) is str:
        return cast(str, tag)
    else:
        sensor_tag = cast(SensorTag, tag)
        return {
            "name": sensor_tag.name,
            "asset": sensor_tag.asset,
        }


class SensorTagNormalizationError(ValueError):
    """Error indicating that something went wrong normalizing a sensor tag"""

    pass



# -*- coding: utf-8 -*-

from abc import ABCMeta, abstractmethod
import logging
from typing import Union, Dict, Any, Tuple

import pandas as pd
import numpy as np
import xarray as xr

from .dataset import _get_dataset


logger = logging.getLogger(__name__)


class ConfigurationError(Exception):
    pass


class GordoBaseDataset(metaclass=ABCMeta):
    def __init__(self):
        self._metadata: Dict[Any, Any] = dict()
        # provided by @capture_args on child's __init__
        if not hasattr(self, "_params"):
            self._params = dict()

    @abstractmethod
    def get_data(
        self,
    ) -> Tuple[
        Union[np.ndarray, pd.DataFrame, xr.DataArray],
        Union[np.ndarray, pd.DataFrame, xr.DataArray],
    ]:
        """
        Return X, y data as numpy or pandas' dataframes given current state
        """

    def get_client_data(
        self, build_dataset_metadata: dict
    ) -> Tuple[
        Union[np.ndarray, pd.DataFrame, xr.DataArray],
        Union[np.ndarray, pd.DataFrame, xr.DataArray],
    ]:
        # TODO better docstring
        return self.get_data()

    def to_dict(self) -> dict:
        """
        Serialize this object into a dict representation, which can be used to
        initialize a new object using :func:`~GordoBaseDataset.from_dict`

        Returns
        -------
        dict
        """
        if not hasattr(self, "_params"):
            raise AttributeError(
                "Failed to lookup init parameters, ensure the "
                "object's __init__ is decorated with 'capture_args'"
            )
        # Update dict with the class
        params = self._params
        params_type = ""
        if hasattr(self, "__module__"):
            # Keep back-compatibility
            if self.__module__ != "gordo_dataset.datasets":
                params_type = self.__module__ + "."
        else:
            print("Not found __module__")
        params_type += self.__class__.__name__
        params["type"] = params_type
        for key, value in params.items():
            if hasattr(value, "to_dict"):
                params[key] = value.to_dict()
        return params

    @classmethod
    def from_dict(cls, config: Dict[str, Any]) -> "GordoBaseDataset":
        """
        Construct the dataset using a config from :func:`~GordoBaseDataset.to_dict`
        """
        return _get_dataset(config)

    def get_metadata(self):
        """
        Get metadata about the current state of the dataset
        """
        return dict()



import os

from typing import Optional
from abc import ABCMeta, abstractmethod
from collections import defaultdict

from gordo_dataset.file_system import FileSystem
from gordo_dataset.file_system.adl1 import ADLGen1FileSystem
from gordo_dataset.file_system.adl2 import ADLGen2FileSystem
from gordo_dataset.file_system.azure import ADLSecret
from gordo_dataset.exceptions import ConfigException


class ADLSecretsLoader(metaclass=ABCMeta):
    """
    Base class for ``ADLSecret`` loader.  Such class can be used for loading
    ``ADLSecret`` from different kinds of sources: files, environment variables, etc.
    """

    @abstractmethod
    def get_secret(self, storage_type: str, storage_name: str) -> Optional[ADLSecret]:
        ...


class ADLEnvSecretsLoader(ADLSecretsLoader):
    """
    Loading ``ADLSecret`` from environment variables.

    Examples
    --------
    >>> secrets_loader = ADLEnvSecretsLoader().from_env("fs", "storage", "TEST_ENV_VAR")
    >>> secrets_loader.get_secret("fs", "storage")
    """

    def __init__(self):
        self._secrets_envs = defaultdict(dict)

    def from_env(self, storage_type: str, storage_name: str, env_var: str):
        self._secrets_envs[storage_type][storage_name] = env_var
        return self

    def get_secret(self, storage_type: str, storage_name: str) -> Optional[ADLSecret]:
        if storage_type not in self._secrets_envs:
            raise ConfigException("Unknown storage type '%s'" % storage_type)
        if storage_name not in self._secrets_envs[storage_type]:
            raise ConfigException(
                "Unknown storage name '%s' for type '%s'" % (storage_type, storage_name)
            )
        env_var_name = self._secrets_envs[storage_type][storage_name]
        env_var = os.environ.get(env_var_name)
        if not env_var:
            return None
        data = env_var.split(":")
        if len(data) != 3:
            raise ValueError(
                "Environment variable %s has %d fields, but 3 is required"
                % (env_var_name, len(data))
            )
        tenant_id, client_id, client_secret = data
        return ADLSecret(tenant_id, client_id, client_secret)



import logging

from datetime import datetime
from io import TextIOWrapper
from azure.datalake.store import core, lib
from typing import Optional, Iterable, IO, Tuple, cast

from gordo_dataset.exceptions import ConfigException

from .base import FileSystem, FileInfo, FileType
from .azure import ADLSecret

logger = logging.getLogger(__name__)


def time_from_info(info: dict, time_key: str) -> Optional[datetime]:
    if time_key in info:
        unix_timestamp = info[time_key]
        return datetime.utcfromtimestamp(unix_timestamp / 1000)
    else:
        return None


class ADLGen1FileSystem(FileSystem):
    @classmethod
    def create_from_env(
        cls,
        store_name: str,
        interactive: bool = False,
        adl_secret: Optional[ADLSecret] = None,
    ) -> "ADLGen1FileSystem":
        """
        Creates ADL Gen1 file system client.

        Parameters
        ----------
        store_name: str
            Name of datalake store.
        interactive: bool
            If true then use interactive authentication
        adl_secret: ADLSecret
            Azure authentication information

        Returns
        -------
        ADLGen1FileSystem
        """

        if interactive:
            logger.info("Attempting to use interactive azure authentication")
            token = lib.auth()
        else:
            if type(adl_secret) is not ADLSecret:
                raise ConfigException(
                    "Unsupported type for adl_secret '%s'" % type(adl_secret)
                )
            adl_secret = cast(ADLSecret, adl_secret)
            logger.info("Attempting to use datalake service authentication")
            token = lib.auth(
                tenant_id=adl_secret.tenant_id,
                client_id=adl_secret.client_id,
                client_secret=adl_secret.client_secret,
                resource="https://datalake.azure.net/",
            )

        adl_client = core.AzureDLFileSystem(token, store_name=store_name)
        return cls(adl_client, store_name)

    def __init__(self, adl_client: core.AzureDLFileSystem, store_name: str):
        self.adl_client = adl_client
        self.store_name = store_name

    @property
    def name(self):
        return self.store_name

    def open(self, path: str, mode: str = "r") -> IO:
        for m in mode:
            if m not in "rb":
                raise ValueError("Unsupported file open mode '%s'" % m)
        wrap_as_text = False
        if "b" not in mode:
            wrap_as_text = True
            mode += "b"
        fd = self.adl_client.open(path, mode=mode)
        return TextIOWrapper(fd) if wrap_as_text else fd

    def exists(self, path: str) -> bool:
        return self.adl_client.exists(path)

    def isfile(self, path: str) -> bool:
        adl_client = self.adl_client
        if adl_client.exists(path):
            info = adl_client.info(path)
            return info["type"] == "FILE"
        else:
            return False

    def isdir(self, path: str) -> bool:
        adl_client = self.adl_client
        if adl_client.exists(path):
            info = adl_client.info(path)
            return info["type"] == "DIRECTORY"
        else:
            return False

    def info(self, path: str) -> FileInfo:
        info = self.adl_client.info(path)
        return self.prepare_info(info)

    @staticmethod
    def prepare_info(info: dict) -> FileInfo:
        if info["type"] == "FILE":
            file_type = FileType.FILE
        elif info["type"] == "DIRECTORY":
            file_type = FileType.DIRECTORY
        else:
            raise ValueError("Unsupported file type '%s'" % info["type"])
        return FileInfo(
            file_type,
            info.get("length", 0),
            access_time=time_from_info(info, "accessTime"),
            modify_time=time_from_info(info, "modificationTime"),
        )

    def ls(
        self, path: str, with_info: bool = True
    ) -> Iterable[Tuple[str, Optional[FileInfo]]]:
        for info in self.adl_client.ls(path, detail=with_info):
            file_path = info["name"] if with_info else info
            file_info = self.prepare_info(info) if with_info else None
            yield file_path, file_info

    def walk(
        self, base_path: str, with_info: bool = True
    ) -> Iterable[Tuple[str, Optional[FileInfo]]]:
        child_directories = []

        for info in self.adl_client.ls(base_path, detail=True):
            file_path = info["name"]
            file_info = self.prepare_info(info) if with_info else None
            if info["type"] == "DIRECTORY":
                child_directories.append(file_path)
            yield file_path, file_info

        for child_directory in child_directories:
            for tup in self.walk(child_directory, with_info=with_info):
                yield tup



import posixpath

from typing import Optional, Iterable, IO, Tuple
from abc import ABCMeta, abstractmethod
from dataclasses import dataclass
from datetime import datetime

from enum import Enum


class FileType(Enum):
    DIRECTORY = 1
    FILE = 2


@dataclass(frozen=True)
class FileInfo:
    file_type: FileType
    size: int
    access_time: Optional[datetime] = None
    modify_time: Optional[datetime] = None
    create_time: Optional[datetime] = None

    def isfile(self) -> bool:
        return self.file_type == FileType.FILE

    def isdir(self) -> bool:
        return self.file_type == FileType.DIRECTORY


def default_join(*p) -> str:
    if p and not p[-1]:
        p = p[:-1]
    if not p:
        return ""
    return posixpath.join(*p)


class FileSystem(metaclass=ABCMeta):
    @property
    @abstractmethod
    def name(self) -> str:
        ...

    @abstractmethod
    def open(self, path: str, mode: str = "r") -> IO:
        """
        Open a file

        Parameters
        ----------
        path
            Path to the file
        mode
            Required support at least of `r`,`b` modes
        Returns
        -------
        IOBase
        """
        ...

    @abstractmethod
    def exists(self, path: str) -> bool:
        ...

    @abstractmethod
    def isfile(self, path: str) -> bool:
        ...

    @abstractmethod
    def isdir(self, path: str) -> bool:
        ...

    @abstractmethod
    def info(self, path: str) -> FileInfo:
        ...

    @abstractmethod
    def ls(
        self, path: str, with_info: bool = True
    ) -> Iterable[Tuple[str, Optional[FileInfo]]]:
        ...

    @abstractmethod
    def walk(
        self, base_path: str, with_info: bool = True
    ) -> Iterable[Tuple[str, Optional[FileInfo]]]:
        ...

    def join(self, *p) -> str:
        return default_join(*p)

    def split(self, p) -> Tuple[str, str]:
        return posixpath.split(p)



import logging

import fnmatch

from gordo_dataset.file_system import FileSystem

from dataclasses import dataclass
from gordo_dataset.data_provider.partition import Partition, MonthPartition
from gordo_dataset.file_system.base import FileInfo
from typing import List, Iterable, Tuple, cast

from .measurement_mapper import Measurement
from .utils import build_dir_path

logger = logging.getLogger(__name__)


@dataclass(frozen=True)
class FileLocation:
    measurement: Measurement
    partition: Partition
    path: str


class NesLookup:
    def __init__(self, storage: FileSystem, base_dir: str):
        """
        Parameters
        ----------
        storage: FileSystem - The file system to find partitions files
        base_dir: str
        """
        self.storage = storage
        self.base_dir = base_dir

    def files_lookup(
        self, measurements: Iterable[Measurement], partitions: Iterable[Partition]
    ) -> Iterable[FileLocation]:
        """
        Finding partitions files in the directory
        """
        storage = self.storage
        collected_partitions: List[MonthPartition] = []
        for partition in partitions:
            if not isinstance(partition, MonthPartition):
                raise ValueError("NesLooks does not support partition %s" % partition)
            collected_partitions.append(partition)
        for measurement in measurements:
            tag = measurement.tag
            for partition in collected_partitions:
                field_values: List[Tuple[str, str]] = [
                    ("asset", tag.asset),
                    ("measurementId", str(measurement.id)),
                    ("year", str(partition.year)),
                    ("month", str(partition.month)),
                ]
                dir_path = build_dir_path(self.storage, self.base_dir, field_values)
                logger.debug(
                    "Processing measurement %s, partition %s, in path '%s'",
                    measurement,
                    partition,
                    dir_path,
                )
                if not storage.isdir(dir_path):
                    logger.warning("Unable to find directory '%s'", dir_path)
                    continue
                found_one = False
                for file_name, file_info in storage.ls(dir_path):
                    file_info = cast(FileInfo, file_info)
                    if file_info.isfile() and fnmatch.fnmatch(file_name, "*.parquet"):
                        found_one = True
                        file_location = FileLocation(measurement, partition, file_name)
                        yield file_location
                if not found_one:
                    logger.warning("Directory '%s is empty", dir_path)



from typing import List, Union, Dict

from .datasets import TimeSeriesDataset, GordoBaseDataProvider, TagList
from .sensor_tag import SensorTag, SensorTagNormalizationError

from .data_provider.nes_provider import NesDataProvider
from .assets_config import AssetsConfig


class NesTimeSeriesDataset(TimeSeriesDataset):
    @staticmethod
    def create_default_data_provider() -> GordoBaseDataProvider:
        return NesDataProvider()

    @staticmethod
    def tag_normalizer(
        assets_config: AssetsConfig,
        sensors: TagList,
        asset: str = None,
    ) -> List[Union[str, SensorTag]]:
        if asset is None:
            raise SensorTagNormalizationError("asset is empty")
        return TimeSeriesDataset.tag_normalizer(assets_config, sensors, asset=asset)



import functools
import inspect

from collections import namedtuple
import logging
from typing import Iterable, Union, List, Callable, Dict, Optional, Tuple, Any, cast
from datetime import datetime

import pandas as pd
import numpy as np

from influxdb import DataFrameClient, InfluxDBClient
from .exceptions import InsufficientDataError
from .sensor_tag import Tag, extract_tag_name, tag_to_json
from .dataset_metadata import tags_to_json_representation


def capture_args(method: Callable):
    """
    Decorator that captures args and kwargs passed to a given method.
    This assumes the decorated method has a self, which has a dict of
    kwargs assigned as an attribute named _params.

    Parameters
    ----------
    method: Callable
        Some method of an object, with 'self' as the first parameter.

    Returns
    -------
    Any
        Returns whatever the original method would return
    """

    @functools.wraps(method)
    def wrapper(self, *args, **kwargs):
        sig_params = inspect.signature(method).parameters.items()

        # Get the default values for the method signature
        params = {
            param: value.default
            for param, value in sig_params
            if value.default is not inspect.Parameter.empty and param != "self"
        }

        # Update with arg values provided
        arg_map = dict()
        for arg_val, arg_key in zip(
            args, (arg for arg in inspect.getfullargspec(method).args if arg != "self")
        ):
            arg_map[arg_key] = arg_val

        # Update params with args/kwargs provided in the current call
        params.update(arg_map)
        params.update(kwargs)

        self._params = params
        return method(self, *args, **kwargs)

    return wrapper


# Prediction result representation, name=str, predictions=dataframe, error_messages=List[str]
PredictionResult = namedtuple("PredictionResult", "name predictions error_messages")


def _parse_influx_uri(uri: str) -> Tuple[str, str, str, str, str, str]:
    """
    Parse an influx URI

    Parameters
    ----------
    uri: str
        Format: <username>:<password>@<host>:<port>/<optional-path>/<db_name>

    Returns
    -------
    (str, str, str, str, str, str)
        username, password, host, port, path, database
    """
    username, password, host, port, *path, db_name = (
        uri.replace("/", ":").replace("@", ":").split(":")
    )
    path_str = "/".join(path) if path else ""
    return username, password, host, port, path_str, db_name


def influx_client_from_uri(
    uri: str,
    api_key: Optional[str] = None,
    api_key_header: Optional[str] = "Ocp-Apim-Subscription-Key",
    recreate: bool = False,
    dataframe_client: bool = False,
    proxies: Dict[str, str] = {"https": "", "http": ""},
) -> Union[InfluxDBClient, DataFrameClient]:
    """
    Get a InfluxDBClient or DataFrameClient from a SqlAlchemy like URI

    Parameters
    ----------
    uri: str
        Connection string format: <username>:<password>@<host>:<port>/<optional-path>/<db_name>
    api_key: str
        Any api key required for the client connection
    api_key_header: str
        The name of the header the api key should be assigned
    recreate: bool
        Re/create the database named in the URI
    dataframe_client: bool
        Return a DataFrameClient instead of a standard InfluxDBClient
    proxies: dict
        A mapping of any proxies to pass to the influx client

    Returns
    -------
    Union[InfluxDBClient, DataFrameClient]
    """

    username, password, host, port, path, db_name = _parse_influx_uri(uri)

    Client = DataFrameClient if dataframe_client else InfluxDBClient

    client = Client(
        host=host,
        port=port,
        database=db_name,
        username=username,
        password=password,
        path=path,
        ssl=bool(api_key),
        proxies=proxies,
    )
    if api_key:
        client._headers[api_key_header] = api_key
    if recreate:
        client.drop_database(db_name)
        client.create_database(db_name)
    return client


def join_timeseries(
    series_iterable: Iterable[Tuple[pd.Series, Tag]],
    resampling_startpoint: datetime,
    resampling_endpoint: datetime,
    resolution: str,
    aggregation_methods: Union[str, List[str], Callable] = "mean",
    interpolation_method: str = "linear_interpolation",
    interpolation_limit: str = "8H",
) -> Tuple[pd.DataFrame, dict]:
    """

    Parameters
    ----------
    series_iterable: Iterable[pd.Series]
        An iterator supplying series with time index
    resampling_startpoint: datetime.datetime
        The starting point for resampling. Most data frames will not have this
        in their datetime index, and it will be inserted with a NaN as the value.
        The resulting NaNs will be removed, so the only important requirement for this is
        that this resampling_startpoint datetime must be before or equal to the first
        (earliest) datetime in the data to be resampled.
    resampling_endpoint: datetime.datetime
        The end point for resampling. This datetime must be equal to or after the last datetime in the
        data to be resampled.
    resolution: str
        The bucket size for grouping all incoming time data (e.g. "10T")
        Available strings come from https://pandas.pydata.org/pandas-docs/stable/user_guide/timeseries.html#dateoffset-objects
    aggregation_methods: Union[str, List[str], Callable]
        Aggregation method(s) to use for the resampled buckets. If a single
        resample method is provided then the resulting dataframe will have names
        identical to the names of the series it got in. If several
        aggregation-methods are provided then the resulting dataframe will
        have a multi-level column index, with the series-name as the first level,
        and the aggregation method as the second level.
        See :py:func::`pandas.core.resample.Resampler#aggregate` for more
        information on possible aggregation methods.
    interpolation_method: str
        How should missing values be interpolated. Either forward fill (`ffill`) or by linear
        interpolation (default, `linear_interpolation`).
    interpolation_limit: str
        Parameter sets how long from last valid data point values will be interpolated/forward filled.
        Default is eight hours (`8H`).
        If None, all missing values are interpolated/forward filled.

    Returns
    -------
    pd.DataFrame
        A dataframe without NaNs, a common time index, and one column per
        element in the dataframe_generator. If multiple aggregation methods
        are provided then the resulting dataframe will have a multi-level column
        index with series-names as top-level and aggregation-method as second-level.
    dict
        Matadata information

    """
    resampled_series = []
    missing_data_series = []
    metadata: Dict[str, Any] = dict()

    tags: List[Tag] = []
    for series, tag in series_iterable:
        series_name = extract_tag_name(tag)
        tags.append(tag)
        metadata[series_name] = dict(original_length=len(series))
        try:
            resampled = _resample(
                series,
                resampling_startpoint=resampling_startpoint,
                resampling_endpoint=resampling_endpoint,
                resolution=resolution,
                aggregation_methods=aggregation_methods,
                interpolation_method=interpolation_method,
                interpolation_limit=interpolation_limit,
            )
        except IndexError:
            missing_data_series.append(series_name)
        else:
            resampled_series.append(resampled)
            metadata[series_name].update(dict(resampled_length=len(resampled)))
            if not resampled.index.empty:
                first_timestamp = resampled.index.min()
                last_timestamp = resampled.index.max()
                gaps_df = _find_gaps(
                    resampled,
                    resolution=resolution,
                    resampling_startpoint=resampling_startpoint,
                    resampling_endpoint=resampling_endpoint,
                )
                metadata[series_name].update(
                    dict(
                        first_timestamp=first_timestamp.timestamp(),
                        last_timestamp=last_timestamp.timestamp(),
                        gaps=_gaps_df_to_dict(gaps_df),
                    )
                )
    if missing_data_series:
        raise InsufficientDataError(
            f"The following features are missing data: {missing_data_series}"
        )

    joined_df = pd.concat(resampled_series, axis=1, join="inner")

    # Before returning, delete all rows with NaN, they were introduced by the
    # insertion of NaNs in the beginning of all timeseries
    dropped_na = joined_df.dropna()

    metadata["aggregate_metadata"] = dict(
        joined_length=len(joined_df), dropped_na_length=len(dropped_na)
    )
    metadata["tags"] = tags_to_json_representation(tags)
    return dropped_na, metadata


def _resample(
    series: pd.Series,
    resampling_startpoint: datetime,
    resampling_endpoint: datetime,
    resolution: str,
    aggregation_methods: Union[str, List[str], Callable] = "mean",
    interpolation_method: str = "linear_interpolation",
    interpolation_limit: str = "8H",
):
    """
    Takes a single series and resamples it.
    See :class:`gordo_dataset.base.GordoBaseDataset.join_timeseries`
    """

    startpoint_sametz = resampling_startpoint.astimezone(tz=series.index[0].tzinfo)
    endpoint_sametz = resampling_endpoint.astimezone(tz=series.index[0].tzinfo)

    if series.index[0] > startpoint_sametz:
        # Insert a NaN at the startpoint, to make sure that all resampled
        # indexes are the same. This approach will "pad" most frames with
        # NaNs, that will be removed at the end.
        startpoint = pd.Series([np.NaN], index=[startpoint_sametz], name=series.name)
        series = startpoint.append(series)
        logging.debug(f"Appending NaN to {series.name} " f"at time {startpoint_sametz}")

    elif series.index[0] < resampling_startpoint:
        msg = (
            f"Error - for {series.name}, first timestamp "
            f"{series.index[0]} is before the resampling start point "
            f"{startpoint_sametz}"
        )
        logging.error(msg)
        raise RuntimeError(msg)

    if series.index[-1] < endpoint_sametz:
        endpoint = pd.Series([np.NaN], index=[endpoint_sametz], name=series.name)
        series = series.append(endpoint)
        logging.debug(f"Appending NaN to {series.name} " f"at time {endpoint_sametz}")
    elif series.index[-1] > endpoint_sametz:
        msg = (
            f"Error - for {series.name}, last timestamp "
            f"{series.index[-1]} is later than the resampling end point "
            f"{endpoint_sametz}"
        )
        logging.error(msg)
        raise RuntimeError(msg)

    logging.debug("Head (3) and tail(3) of dataframe to be resampled:")
    logging.debug(series.head(3))
    logging.debug(series.tail(3))

    resampled = series.resample(resolution, label="left").agg(aggregation_methods)
    # If several aggregation methods are provided, agg returns a dataframe
    # instead of a series. In this dataframe the column names are the
    # aggregation methods, like "max" and "mean", so we have to make a
    # multi-index with the series-name as the top-level and the
    # aggregation-method as the lower-level index.
    # For backwards-compatibility we *dont* return a multi-level index
    # when we have a single resampling method.
    if isinstance(resampled, pd.DataFrame):  # Several aggregation methods provided
        resampled.columns = pd.MultiIndex.from_product(
            [[series.name], resampled.columns], names=["tag", "aggregation_method"]
        )

    if interpolation_method not in ["linear_interpolation", "ffill"]:
        raise ValueError(
            "Interpolation method should be either linear_interpolation of ffill"
        )

    if interpolation_limit is not None:
        limit = int(
            pd.Timedelta(interpolation_limit).total_seconds()
            / pd.Timedelta(resolution).total_seconds()
        )

        if limit <= 0:
            raise ValueError("Interpolation limit must be larger than given resolution")
    else:
        limit = None

    if interpolation_method == "linear_interpolation":
        return resampled.interpolate(limit=limit).dropna()
    else:
        return resampled.fillna(method=interpolation_method, limit=limit).dropna()


def _to_timestamps(column: List[pd.Timestamp]):
    return [v.timestamp() for v in column]


def _gaps_df_to_dict(df: pd.DataFrame) -> dict:
    """
    Extract metadata information from `_find_gaps` DataFrame

    Parameters
    ----------
    df: pd.DataFrame
        Contains columns: start, end
    """
    if df.empty:
        return {}
    df_dict = df.to_dict("list")
    return {
        "start": _to_timestamps(df_dict["start"]),
        "end": _to_timestamps(df_dict["end"]),
    }


def _find_gaps(
    values: Union[pd.Series, pd.DataFrame],
    resolution: str,
    resampling_startpoint: datetime,
    resampling_endpoint: datetime,
) -> pd.DataFrame:
    """Find the gaps in a series's index and create a dataframe to store them, with
    columns: start, end
    """

    index = values.index
    td = pd.to_timedelta(resolution)
    s = index.to_series()
    # Add start/end points to catch heading/trailing gaps
    startpoint_sametz = resampling_startpoint.astimezone(tz=index[0].tzinfo)
    endpoint_sametz = resampling_endpoint.astimezone(tz=index[0].tzinfo)
    start_end = pd.Series(
        [startpoint_sametz, endpoint_sametz], index=[startpoint_sametz, endpoint_sametz]
    )
    s = s.append(start_end).sort_values()
    diff_ = s.diff()[1:]  # Ignore first NaN
    g = diff_[diff_ > td]

    # Output DataFrame
    gaps = pd.DataFrame(g, columns=["duration"])
    gaps["start"] = gaps.index - gaps["duration"]
    gaps = gaps.rename_axis("end").reset_index()
    return gaps[["start", "end"]]



import logging
import pandas as pd
import pyarrow.parquet as pq
import fnmatch

from typing import List, Dict, cast
from abc import abstractmethod, ABCMeta
from gordo_dataset.sensor_tag import SensorTag
from gordo_dataset.data_provider.storages import FileSystem
from gordo_dataset.file_system.base import FileInfo
from dataclasses import dataclass
from collections import defaultdict

from .utils import build_dir_path

logger = logging.getLogger(__name__)


class MeasurementNotFoundError(Exception):
    def __init__(self, msg: str, sensor_tags: List[SensorTag]):
        self.sensor_tags = sensor_tags
        super().__init__(msg)


@dataclass(frozen=True)
class Measurement:
    id: int
    tag: SensorTag


class MeasurementMapper(metaclass=ABCMeta):
    """
    Find measurement-ids from the sensor tag name
    """

    @abstractmethod
    def get_measurements(self, sensor_tags: List[SensorTag]) -> List[Measurement]:
        ...


class DirectMeasurementMapper(MeasurementMapper):
    def get_measurements(self, sensor_tags: List[SensorTag]) -> List[Measurement]:
        measurements = []
        for sensor_tag in sensor_tags:
            measurements.append(Measurement(sensor_tag.name, sensor_tag))
        return measurements


class DataLakeMeasurementMapper(MeasurementMapper):
    def __init__(self, storage: FileSystem, base_dir: str):
        self.storage = storage
        self.base_dir = base_dir

    def read_parquet(self, path: str) -> pd.Series:
        with self.storage.open(path, "rb") as f:
            table = pq.read_table(f, columns=["measurementId", "measurementName"])
            df = table.to_pandas()
            df = df.set_index("measurementName")
        return df["measurementId"]

    def find_measurements_in_dir(
        self, dir_path: str, sensor_tags: List[SensorTag]
    ) -> List[Measurement]:
        storage = self.storage
        name_ids: Dict[str, int] = {}
        for file_path, file_info in storage.ls(dir_path):
            file_info = cast(FileInfo, file_info)
            if file_info.isfile() and fnmatch.fnmatch(file_path, "*.parquet"):
                s = self.read_parquet(file_path)
                found_name_ids = {}
                for sensor_tag in sensor_tags:
                    name = sensor_tag.name
                    try:
                        found_name_ids[name] = s.loc[name]
                    except KeyError:
                        pass
                logger.debug(
                    "Read parquet '%s'. Find %d measurements from %d",
                    file_path,
                    len(found_name_ids),
                    len(sensor_tags),
                )
                for name in found_name_ids.keys():
                    if name in name_ids:
                        raise ValueError(
                            "measurementName='%s' duplicate in the file '%s'"
                            % (name, file_path)
                        )
                name_ids.update(found_name_ids)
        measurements = []
        for sensor_tag in sensor_tags:
            name = sensor_tag.name
            if name not in name_ids:
                raise MeasurementNotFoundError(
                    "Unable to find measurement for the tag %s" % (sensor_tag,),
                    [sensor_tag],
                )
            measurements.append(Measurement(name_ids[name], sensor_tag))
        return measurements

    def get_measurements(self, sensor_tags: List[SensorTag]) -> List[Measurement]:
        storage, base_dir = self.storage, self.base_dir
        by_assets = defaultdict(list)
        for sensor_tag in sensor_tags:
            if sensor_tag.asset is None:
                raise ValueError("asset is empty for %s" % (str(sensor_tag),))
            by_assets[sensor_tag.asset].append(sensor_tag)
        measurements: List[Measurement] = []
        for asset, asset_sensor_tags in by_assets.items():
            dir_path = build_dir_path(storage, base_dir, [("asset", asset)])
            if not storage.isdir(dir_path):
                msg = "Unable to find asset '%s' directory '%s'" % (asset, dir_path)
                raise MeasurementNotFoundError(msg, asset_sensor_tags)
            asset_measurements = self.find_measurements_in_dir(
                dir_path, asset_sensor_tags
            )
            measurements.extend(asset_measurements)
        return measurements



import logging
import threading
import math

from azure.core.exceptions import ResourceNotFoundError
from azure.core.pipeline.transport import RequestsTransport
from azure.storage.filedatalake import (
    DataLakeServiceClient,
    FileSystemClient,
    PathProperties,
)
from azure.identity import (
    ClientSecretCredential,
    AzureCliCredential,
    InteractiveBrowserCredential,
    DefaultAzureCredential,
    ChainedTokenCredential,
)
from typing import Optional, Any, IO, Iterable, cast, Tuple
from io import BytesIO, TextIOWrapper

from .base import FileSystem, FileInfo, FileType
from .azure import ADLSecret

logger = logging.getLogger(__name__)


class ThreadLocalRequestTransport(RequestsTransport):
    def __init__(self, **kwargs):
        self.local = threading.local()
        super().__init__(**kwargs)

    @property
    def session(self):
        return self.local.__dict__.get("session", None)

    @session.setter
    def session(self, session):
        self.local.session = session


class ADLGen2FileSystem(FileSystem):
    @classmethod
    def create_from_env(
        cls,
        account_name: str,
        file_system_name: str,
        interactive: bool = False,
        adl_secret: Optional[ADLSecret] = None,
        **kwargs,
    ) -> "ADLGen2FileSystem":
        """
        Creates ADL Gen2 file system client.

        Parameters
        ----------
        account_name: str
            Azure account name
        file_system_name: str
            Container name
        interactive: bool
            If true then use interactive authentication
        adl_secret: ADLSecret
            Azure authentication information

        Returns
        -------
        ADLGen2FileSystem
        """
        if interactive:
            logger.info("Attempting to use interactive azure authentication")
            credential = ChainedTokenCredential(
                AzureCliCredential(), InteractiveBrowserCredential()
            )
        elif type(adl_secret) is ADLSecret:
            adl_secret = cast(ADLSecret, adl_secret)
            logger.info("Attempting to use datalake service authentication")
            credential = ClientSecretCredential(
                tenant_id=adl_secret.tenant_id,
                client_id=adl_secret.client_id,
                client_secret=adl_secret.client_secret,
            )
        else:
            credential = DefaultAzureCredential()
        return cls.create_from_credential(
            account_name, file_system_name, credential, **kwargs
        )

    @classmethod
    def create_from_credential(
        cls,
        account_name: str,
        file_system_name: str,
        credential: Any,
        use_thread_local_transport: bool = True,
        **kwargs,
    ) -> "ADLGen2FileSystem":
        """
        Creates ADL Gen2 file system client.

        Parameters
        ----------
        account_name: str
            Azure account name
        file_system_name: str
            Container name
        credential: object
            azure.identity credential
        use_thread_local_transport: bool
            Use ``ThreadLocalRequestTransport`` as HTTP transport

        Returns
        -------
        ADLGen2FileSystem
        """
        client_kwargs = {}
        if use_thread_local_transport:
            client_kwargs["transport"] = ThreadLocalRequestTransport()
        service_client = DataLakeServiceClient(
            account_url="https://%s.dfs.core.windows.net" % account_name,
            credential=credential,
            **client_kwargs,
        )
        file_system_client = service_client.get_file_system_client(
            file_system=file_system_name
        )
        return cls(file_system_client, account_name, file_system_name, **kwargs)

    def __init__(
        self,
        file_system_client: FileSystemClient,
        account_name: str,
        file_system_name: str,
        thread_chunk_size: int = 50 * (10 ** 6),  # 50 Mb
        max_threads_count: int = 20,
    ):
        self.file_system_client = file_system_client
        self.account_name = account_name
        self.file_system_name = file_system_name
        self.thread_chunk_size = thread_chunk_size
        self.max_threads_count = max_threads_count

    @staticmethod
    def get_max_concurrency(
        file_size: int, thread_chunk_size: int, max_threads_count: int
    ) -> int:
        if thread_chunk_size:
            threads_count = math.ceil(file_size / thread_chunk_size)
            return min(threads_count, max_threads_count)
        else:
            return max_threads_count

    @property
    def name(self):
        return self.file_system_name + "@" + self.account_name

    def open(self, path: str, mode: str = "r") -> IO:
        for m in mode:
            if m not in "rb":
                raise ValueError("Unsupported file open mode '%s'" % m)
        wrap_as_text = "b" not in mode
        info = self.info(path)
        if info.size:
            max_concurrency = self.get_max_concurrency(
                info.size, self.thread_chunk_size, self.max_threads_count
            )
            file_client = self.file_system_client.get_file_client(path)
            downloader = file_client.download_file(max_concurrency=max_concurrency)
            stream = BytesIO(downloader.readall())
        else:
            stream = BytesIO(b"")
        fd = cast(IO, TextIOWrapper(stream) if wrap_as_text else stream)
        return fd

    def exists(self, path: str) -> bool:
        try:
            self.info(path)
            return True
        except FileNotFoundError:
            return False

    def isfile(self, path: str) -> bool:
        try:
            info = self.info(path)
            return info.file_type == FileType.FILE
        except FileNotFoundError:
            return False

    def isdir(self, path: str) -> bool:
        try:
            info = self.info(path)
            return info.file_type == FileType.DIRECTORY
        except FileNotFoundError:
            return False

    def info(self, path: str) -> FileInfo:
        file_client = self.file_system_client.get_file_client(path)
        try:
            properties = file_client.get_file_properties()
        except ResourceNotFoundError:
            raise FileNotFoundError(path)
        except Exception as e:
            logger.debug("Exception %s(%s)", e.__class__.__name__, path)
            raise
        content_settings = properties["content_settings"]
        if content_settings.get("content_type", None):
            file_type = FileType.FILE
        else:
            file_type = FileType.DIRECTORY
        return FileInfo(
            file_type,
            properties.get("size", 0),
            modify_time=properties["last_modified"],
            create_time=properties["creation_time"],
        )

    @staticmethod
    def path_properties_to_info(path_properties: PathProperties) -> FileInfo:
        if path_properties.is_directory:
            file_type = FileType.DIRECTORY
        else:
            file_type = FileType.FILE
        size = path_properties.content_length if path_properties.content_length else 0
        # TODO taking into account attribute path_properties.last_modified. Example: 'Mon, 07 Sep 2020 12:01:14 GMT'
        return FileInfo(file_type, size)

    @staticmethod
    def _handle_attribute_error_bug(e: AttributeError):
        if e.__cause__ is not None and hasattr(e.__cause__, "response"):
            response = getattr(e.__cause__, "response")
            logger.error(
                "Handling azure.storage.filedatalake AttributeError bug. Response: %s",
                response,
            )
        raise e

    def ls(
        self, path: str, with_info: bool = True
    ) -> Iterable[Tuple[str, Optional[FileInfo]]]:
        dir_iterator = self.file_system_client.get_paths(path, recursive=False)

        try:
            for properties in dir_iterator:
                file_info = (
                    self.path_properties_to_info(properties) if with_info else None
                )
                yield properties.name, file_info
        except AttributeError as e:
            self._handle_attribute_error_bug(e)

    def walk(
        self, base_path: str, with_info: bool = True
    ) -> Iterable[Tuple[str, Optional[FileInfo]]]:
        dir_iterator = self.file_system_client.get_paths(base_path)

        try:
            for properties in dir_iterator:
                file_info = (
                    self.path_properties_to_info(properties) if with_info else None
                )
                yield properties.name, file_info
        except AttributeError as e:
            self._handle_attribute_error_bug(e)



from urllib.parse import quote

from typing import Iterable, Tuple, Union
from gordo_dataset.data_provider.storages import FileSystem


def partition_dir_name(field: str, value: str):
    return field + "=" + quote(value, safe=" ")


def build_dir_path(
    storage: FileSystem, base_dir: str, field_values: Iterable[Tuple[str, str]]
) -> str:
    dir_path = base_dir
    for field, value in field_values:
        dir_path = storage.join(dir_path, partition_dir_name(field, value))
    return dir_path



# -*- coding: utf-8 -*-
import importlib
from .exceptions import ConfigException


def _get_dataset(config):
    """
    Return a GordoBaseDataSet object of a certain type, given a config dict
    """
    dataset_config = dict(config)
    kind = dataset_config.pop("type", "")
    if "." in kind:
        module_name, class_name = kind.rsplit(".", 1)
        # TODO validate module_name
        module = importlib.import_module(module_name)
        if not hasattr(module, class_name):
            raise ConfigException(
                "Unable to find class %s in module %s" % (module_name, class_name)
            )
        Dataset = getattr(module, class_name)
    else:
        import gordo_dataset.datasets as datasets

        if not kind:
            kind = "TimeSeriesDataset"
        if not hasattr(datasets, kind):
            raise ConfigException(
                "Unable to find class %s in module gordo_dataset.datasets" % kind
            )
        Dataset = getattr(datasets, kind)
    if Dataset is None:
        raise ConfigException(f'Dataset type "{kind}" is not supported!')

    return Dataset(**dataset_config)



from dataclasses import dataclass
from enum import Enum
from typing import Union, Iterable, Optional
from datetime import datetime


@dataclass(frozen=True)
class YearPartition:
    year: int

    def __lt__(self, other):
        if not isinstance(other, YearPartition):
            raise NotImplementedError()
        return self.year < other.year


@dataclass(frozen=True)
class MonthPartition:
    year: int
    month: int

    def __lt__(self, other):
        if not isinstance(other, MonthPartition):
            raise NotImplementedError()
        if self.year == other.year:
            return self.month < other.month
        return self.year < other.year


Partition = Union[YearPartition, MonthPartition]


class PartitionBy(Enum):
    YEAR = "year"
    MONTH = "month"

    @classmethod
    def find_by_name(cls, name) -> Optional["PartitionBy"]:
        for enum in cls:
            if enum.value == name:
                return enum
        return None


def split_by_partitions(
    partition_by: PartitionBy, start_period: datetime, end_period: datetime
) -> Iterable[Partition]:
    if start_period > end_period:
        message = "start_period bigger then end_period."
        message += "'%s' > '%s'" % (start_period.isoformat(), end_period.isoformat())
        raise ValueError(message)
    if partition_by is PartitionBy.YEAR:
        for year in range(start_period.year, end_period.year + 1):
            yield YearPartition(year)
    elif partition_by is PartitionBy.MONTH:
        for year in range(start_period.year, end_period.year + 1):
            for month in range(1, 12 + 1):
                if (year == start_period.year and month < start_period.month) or (
                    year == end_period.year and month > end_period.month
                ):
                    continue
                yield MonthPartition(year, month)
    else:
        raise ValueError("Unknown partition_by type %s" % partition_by)



import pkg_resources

from io import TextIOWrapper
from .assets_config import AssetsConfig

LEGACY_DEFAULT_STORAGE = "dls@omniadlseun"


def load_assets_config() -> AssetsConfig:
    # TODO migrate to importlib.resources after deprecating python 3.6 support
    with pkg_resources.resource_stream(
        "gordo_dataset.resources", "assets_config.yaml"
    ) as f:
        assets_config = AssetsConfig.load_from_yaml(
            TextIOWrapper(f), "assets_config.yaml"
        )
    return assets_config



import yaml
import re

from typing import Optional, Dict, IO, List, Pattern, Tuple, Set, Any, Union
from dataclasses import dataclass

from .exceptions import ConfigException
from .file_system.base import FileSystem

from pydantic import BaseModel, ValidationError
from collections import defaultdict


class AssetsModel(BaseModel):
    aliases: Dict[str, str]


class PathAsset(BaseModel):
    path: str
    asset: str


class StorageAsset(BaseModel):
    name: str
    tag_pattern: Optional[str]
    path: str


class StorageItems(BaseModel):
    base_dir: str
    assets: List[StorageAsset]


class ConfigModel(BaseModel):
    assets: Optional[AssetsModel] = None
    storages: Dict[str, List[StorageItems]]


@dataclass(frozen=True)
class PathSpec:
    asset: str
    base_dir: str
    path: str
    tag_pattern: Optional[Pattern] = None

    def full_path(self, fs: FileSystem) -> str:
        return fs.join(self.base_dir, self.path)


def pydantic_loc_to_json_path(loc: Tuple[Union[int, str], ...]) -> str:
    if len(loc):
        json_paths = []
        for v in loc:
            format_str = "[%d]" if type(v) is int else ".%s"
            json_paths.append(format_str % v)
        json_path = "".join(json_paths)
        return json_path if json_path[0] == "." else "." + json_path
    return ""


def exception_message(message: str, file_path: Optional[str] = None) -> str:
    if file_path:
        return message + ". Config path: %s" % file_path
    else:
        return message


def validation_error_exception_message(errors: List[Dict[str, Any]]) -> str:
    error_messages = []
    for error in errors:
        error_messages.append(
            "%s in '%s'" % (error["msg"], pydantic_loc_to_json_path(error["loc"]))
        )
    message = "Validation error%s: %s" % (
        "" if len(errors) <= 1 else "s",
        "; ".join(error_messages),
    )
    return message


WithPattern = Dict[str, List[Tuple[Pattern, PathSpec]]]
WithoutPattern = Dict[str, List[PathSpec]]
ByAssets = Dict[str, Dict[str, List[PathSpec]]]


class AssetsConfig:
    @classmethod
    def load_from_yaml(
        cls, f: IO[str], file_path: Optional[str] = None
    ) -> "AssetsConfig":
        """
        Loading AssetsConfig from YAML file

        Parameters
        ----------
        f: IO[str]
            File object
        file_path
            Source file path. Using only for exception messages

        Returns
        -------

        """
        raw_config = yaml.safe_load(f)
        return cls.load(raw_config, file_path=file_path)

    @classmethod
    def load(cls, raw_config: dict, file_path: Optional[str] = None) -> "AssetsConfig":
        """
        Loading AssetsConfig from a dictionary. See ``load_from_yaml`` method for loading from YAML file

        Examples
        --------
        >>> raw_config = {'storages': {'adlstore': [{'assets': [{'name': 'asset1',
        ...                             'tag_pattern': '^tag1',
        ...                             'path': 'path/to/asset1'},
        ...                            {'name': 'asset2',
        ...                             'path': 'path/to/asset2'}],
        ...                  'base_dir': '/ncs_data'
        ...                  }]}}
        >>> config = AssetsConfig.load(raw_config)
        >>> config.get_paths("adlstore", "tag11")
        [PathSpec(asset='asset1', base_dir='/ncs_data', path='path/to/asset1', tag_pattern=re.compile('^tag1', re.IGNORECASE))]
        >>> config.get_paths("adlstore", "tag22", "asset2")
        [PathSpec(asset='asset2', base_dir='/ncs_data', path='path/to/asset2', tag_pattern=None)]

        Parameters
        ----------
        raw_config: dict
            Config source
        file_path
            Source file path. Using only for exception messages


        Returns
        -------
        AssetsConfig

        """
        try:
            config = ConfigModel(**raw_config)
        except ValidationError as e:
            message = validation_error_exception_message(e.errors())
            raise ConfigException(exception_message(message, file_path))
        storages = {}
        for storage, storage_items in config.storages.items():
            path_specs: List[PathSpec] = []
            for storage_item in storage_items:
                paths_set: Set[Tuple[str, str]] = set()
                base_dir = storage_item.base_dir
                for storage_asset in storage_item.assets:
                    path = storage_asset.path
                    key = (base_dir, path)
                    if key in paths_set:
                        message = (
                            f"Found duplicate path in storage '{storage}' with base "
                            f"dir '{base_dir}', and path '{path}'"
                        )
                        raise ConfigException(message)
                    name = storage_asset.name
                    tag_pattern = None
                    if storage_asset.tag_pattern:
                        try:
                            tag_pattern = re.compile(
                                storage_asset.tag_pattern, re.IGNORECASE
                            )
                        except re.error as e:
                            error = str(e)
                            message = (
                                f"Unable to compile tag_pattern in storage '{storage}' with base "
                                f"dir '{base_dir}', and path '{path}', because: {error}"
                            )
                            raise ConfigException(message)
                    path_spec = PathSpec(
                        name, base_dir, storage_asset.path, tag_pattern
                    )
                    path_specs.append(path_spec)
                    paths_set.add(key)
            storages[storage] = path_specs
        assets_aliases: Optional[Dict[str, str]] = None
        if config.assets is not None:
            assets_aliases = config.assets.aliases
        return cls(storages, assets_aliases)

    def __init__(
        self,
        storages: Dict[str, List[PathSpec]],
        assets_aliases: Optional[Dict[str, str]] = None,
    ):
        self.storage_names = set(storages.keys())
        self.with_pattern, self.without_pattern, self.by_assets = self.prepare(storages)
        self.assets_aliases = assets_aliases

    @staticmethod
    def prepare(
        storages: Dict[str, List[PathSpec]]
    ) -> Tuple[WithPattern, WithoutPattern, ByAssets]:
        with_pattern: WithPattern = defaultdict(list)
        without_pattern: WithoutPattern = defaultdict(list)
        by_assets: ByAssets = defaultdict(dict)
        for storage, path_specs in storages.items():
            storage_by_asset = by_assets[storage]
            for path_spec in path_specs:
                asset = path_spec.asset
                if asset not in storage_by_asset:
                    storage_by_asset[asset] = []
                storage_by_asset[asset].append(path_spec)
                if path_spec.tag_pattern:
                    with_pattern[storage].append((path_spec.tag_pattern, path_spec))
                else:
                    without_pattern[storage].append(path_spec)
        return with_pattern, without_pattern, by_assets

    def prepare_asset_name(self, asset: str) -> str:
        if self.assets_aliases is not None:
            return self.assets_aliases.get(asset, asset)
        return asset

    def get_paths(
        self,
        storage: str,
        tag: str,
        asset: Optional[str] = None,
    ) -> List[PathSpec]:
        """
        Tries to find tag corresponded path in the config

        Parameters
        ----------
        storage
            Storage name. For Azure Data Lake Gen2 <container name>@<storage account>
        tag
            Tag name
        asset
            Asset name if known. If None will be determined through tag_patterns in the config

        Returns
        -------
        List[PathSpec]

        """
        path_specs: List[PathSpec] = []
        if storage in self.storage_names:
            if asset is not None:
                asset = self.prepare_asset_name(asset)
                if storage in self.by_assets:
                    path_specs = self.by_assets[storage].get(asset, [])
            else:
                if storage in self.with_pattern:
                    for tag_pattern, path_spec in self.with_pattern[storage]:
                        if tag_pattern.match(tag):
                            path_specs.append(path_spec)
                if not path_specs:
                    path_specs = self.without_pattern.get(storage, [])
        return path_specs



from typing import Iterable, Set, Optional, Dict, cast, List

from .assets_config import AssetsConfig
from .resource_assets_config import LEGACY_DEFAULT_STORAGE

from .sensor_tag import (
    tag_to_json,
    extract_tag_name,
    Tag,
    SensorTag,
    normalize_sensor_tag,
    legacy_normalize_sensor_tag,
    load_sensor_tag,
)


def tags_to_json_representation(tags: Iterable[Tag]) -> dict:
    tags_metadata = {}
    for tag in tags:
        tag_name = extract_tag_name(tag)
        json_repr = tag_to_json(tag)
        if type(json_repr) is str:
            continue
        tags_metadata[tag_name] = json_repr
    return tags_metadata


def _get_dataset_meta(build_dataset_metadata: dict) -> Optional[dict]:
    if "dataset_meta" in build_dataset_metadata["dataset"]:
        return build_dataset_metadata["dataset"]["dataset_meta"]
    return None


def _tags_from_build_metadata(build_dataset_metadata: dict) -> Optional[dict]:
    dataset_meta = _get_dataset_meta(build_dataset_metadata)
    if dataset_meta is not None:
        if "tag_loading_metadata" in dataset_meta:
            tag_loading_metadata = dataset_meta["tag_loading_metadata"]
            return tag_loading_metadata.get("tags")
    return None


def _storage_from_build_metadata(build_dataset_metadata: dict) -> Optional[str]:
    dataset_meta = _get_dataset_meta(build_dataset_metadata)
    if dataset_meta is not None:
        if "data_provider" in dataset_meta:
            provider_metadata = dataset_meta["data_provider"]
            return provider_metadata.get("storage_name")
    return None


def _legacy_tags_normalization(
    build_dataset_metadata: dict,
    tag_list: List[str],
    assets_config: AssetsConfig,
    storage: Optional[str] = None,
    asset: Optional[str] = None,
) -> Dict[str, SensorTag]:
    sensor_tags: Dict[str, SensorTag] = {}
    for tag in tag_list:
        curr_assets_config = cast(AssetsConfig, assets_config)
        if storage is None:
            storage = _storage_from_build_metadata(build_dataset_metadata)
        curr_storage: str = storage if storage is not None else LEGACY_DEFAULT_STORAGE
        sensor = normalize_sensor_tag(curr_assets_config, tag, asset=asset)
        sensor_tag = legacy_normalize_sensor_tag(
            curr_assets_config, curr_storage, sensor
        )
        sensor_tags[extract_tag_name(sensor_tag)] = sensor_tag
    return sensor_tags


_list_of_tags_exception_message = (
    "The list of tags should be placed on"
    " dataset.dataset_meta.tag_loading_metadata.tags path"
)


def sensor_tags_from_build_metadata(
    build_dataset_metadata: dict,
    tag_names: Set[str],
    *,
    with_legacy_tag_normalization: bool = True,
    assets_config: Optional[AssetsConfig] = None,
    storage: Optional[str] = None,
    asset: Optional[str] = None,
) -> Dict[str, SensorTag]:
    # TODO better docstring
    # TODO unit-tests
    if with_legacy_tag_normalization:
        if assets_config is None:
            raise ValueError(
                "assets_config should be provided with_legacy_tag_normalization"
            )
    tags_build_metadata = _tags_from_build_metadata(build_dataset_metadata)
    sensor_tags: Dict[str, SensorTag] = {}
    if tags_build_metadata is None:
        if with_legacy_tag_normalization:
            tag_list = list(tag_names)
            return _legacy_tags_normalization(
                build_dataset_metadata,
                tag_list,
                cast(AssetsConfig, assets_config),
                storage=storage,
                asset=asset,
            )
        raise ValueError(
            "Unable to find tags information in build_metadata. "
            + _list_of_tags_exception_message
        )
    for tag_name in tag_names:
        if tag_name not in tags_build_metadata:
            raise ValueError(
                "Unable to find tag '%s' information in build_metadata. "
                + _list_of_tags_exception_message
            )
        tag_metadata = tags_build_metadata[tag_name]
        sensor_tag = load_sensor_tag(tag_metadata)
        sensor_tags[sensor_tag.name] = sensor_tag
    return sensor_tags



import logging
import timeit
import os

from datetime import datetime
from copy import copy

import pandas as pd
import pyarrow.parquet as pq
from concurrent.futures import ThreadPoolExecutor

from gordo_dataset.file_system import FileSystem
from gordo_dataset.utils import capture_args
from gordo_dataset.sensor_tag import SensorTag

from gordo_dataset.data_provider.base import GordoBaseDataProvider
from gordo_dataset.data_provider.storages import create_storage, DEFAULT_STORAGE_TYPE
from gordo_dataset.data_provider.partition import split_by_partitions, PartitionBy
from gordo_dataset.exceptions import ConfigException
from typing import Union, Dict, Optional, Any, List, Iterable, Tuple, TypeVar, cast


from .measurement_mapper import MeasurementMapper, DataLakeMeasurementMapper
from .nes_lookup import NesLookup, FileLocation

logger = logging.getLogger(__name__)

FileLoadingResult = Tuple[FileLocation, pd.Series]

DEFAULT_GOOD_STATUSES = [192]


T = TypeVar("T")


def infinity_iter(v: T) -> Iterable[T]:
    while True:
        yield v


def get_default_base_dir() -> Optional[str]:
    adl_base_dir = os.environ.get("ADL_BASE_DIR")
    if not adl_base_dir:
        adl_base_dir = "raw/data/bazefield wind/omnia_prevent/"
    return adl_base_dir


class NesDataProvider(GordoBaseDataProvider):
    @staticmethod
    def instantiate_storage(
        storage: Optional[Union[FileSystem, Dict[str, Any]]]
    ) -> FileSystem:
        if isinstance(storage, FileSystem):
            return storage
        else:
            kwargs: Dict[str, Any] = copy(storage) if storage else {}
            storage_type = kwargs.pop("type", DEFAULT_STORAGE_TYPE)
            return create_storage(storage_type, **kwargs)

    @staticmethod
    def create_measurement_mapper(
        storage: FileSystem, base_dir: str
    ) -> MeasurementMapper:
        metatable_dir = storage.join(base_dir, "metatable")
        return DataLakeMeasurementMapper(storage, metatable_dir)

    @staticmethod
    def create_nes_mapper(storage: FileSystem, base_dir: str) -> NesLookup:
        data_dir = storage.join(base_dir, "data")
        return NesLookup(storage, data_dir)

    @capture_args
    def __init__(
        self,
        storage: Optional[Union[FileSystem, Dict[str, Any]]] = None,
        measurement_mapper: Optional[MeasurementMapper] = None,
        base_dir: Optional[str] = None,
        good_statuses: Optional[List[int]] = None,
        threads: int = 10,
    ):
        """
        Parameters
        ----------
        storage: Optional[Union[FileSystem, Dict[str, Any]]] - Storage with NES data
        measurement_mapper: Optional[MeasurementMapper] - The measurement mapper. By default it's ``DataLakeMeasurementMapper``
        base_dir: Optional[str] - Directory where are the data and the metatable
        good_statuses: Optional[List[int]] - Filter out all values which is not in this list
        threads: int - Number of threads in the fetcher thread pool

        """
        if not base_dir:
            base_dir = get_default_base_dir()
        self.storage = self.instantiate_storage(storage)
        if not measurement_mapper:
            measurement_mapper = self.create_measurement_mapper(
                self.storage, cast(str, base_dir)
            )
        self.measurement_mapper = measurement_mapper
        self.lookup = self.create_nes_mapper(self.storage, cast(str, base_dir))
        if not good_statuses:
            good_statuses = DEFAULT_GOOD_STATUSES
        self.good_statuses = good_statuses
        self.threads = threads

    def read_parquet(self, path: str, tag_name: str) -> pd.Series:
        with self.storage.open(path, "rb") as f:
            table = pq.read_table(f, columns=["time", "q", "v"])
            df = table.to_pandas()
            if self.good_statuses:
                df = df[df["q"].isin(self.good_statuses)]
            df["time"] = pd.to_datetime(df["time"], utc=True)
            df["v"] = pd.to_numeric(df["v"])
            df = df.set_index("time")
            df = df.rename(columns={"v": tag_name})
        return df[tag_name]

    def load_location(
        self, file_location: FileLocation, dry_run: bool = False
    ) -> pd.Series:
        measurement = file_location.measurement
        tag = measurement.tag
        path = file_location.path
        logger.info(
            "Downloading measurement: %s for partitions: %s from '%s'",
            measurement,
            file_location.partition,
            path,
        )

        info = self.storage.info(path)
        file_size = info.size / (1024 ** 2)
        logger.debug("File size for file '%s': %.2fMB", path, file_size)

        if dry_run:
            # Dry run only, returning empty frame early
            return pd.Series(name=tag.name)

        before_downloading = timeit.default_timer()
        s = self.read_parquet(path, tag.name)
        time_elapsed = timeit.default_timer() - before_downloading
        logger.debug("Done in %.2f sec %s", time_elapsed, path)
        return s

    def _thread_pool_mapper(
        self, file_location: FileLocation, dry_run: bool
    ) -> FileLoadingResult:
        return file_location, self.load_location(file_location, dry_run)

    def load_files(
        self, file_locations: Iterable[FileLocation], dry_run: bool
    ) -> Iterable[FileLoadingResult]:
        if self.threads > 1:
            dry_run_iter = infinity_iter(dry_run)
            with ThreadPoolExecutor(max_workers=self.threads) as executor:
                for result in executor.map(
                    self._thread_pool_mapper, file_locations, dry_run_iter
                ):
                    yield result
        else:
            for file_location in file_locations:
                yield file_location, self.load_location(file_location, dry_run)

    def load_series(
        self,
        train_start_date: datetime,
        train_end_date: datetime,
        tag_list: List[Union[str, SensorTag]],
        dry_run: Optional[bool] = False,
        **kwargs,
    ) -> Iterable[pd.Series]:
        for tag in tag_list:
            if type(tag) is str:
                raise ConfigException(
                    "'%s' tag does not have an asset provided" % cast(str, tag)
                )
        sensor_tag_list = cast(List[SensorTag], tag_list)
        if train_end_date < train_start_date:
            raise ConfigException(
                f"NCS reader called with train_end_date: {train_end_date} before train_start_date: {train_start_date}"
            )
        measurements = self.measurement_mapper.get_measurements(sensor_tag_list)
        partitions = split_by_partitions(
            PartitionBy.MONTH, train_start_date, train_end_date
        )
        file_locations = self.lookup.files_lookup(measurements, partitions)
        series: Dict[str, pd.Series] = {}
        for file_location, s in self.load_files(file_locations, bool(dry_run)):
            name = s.name
            if name not in series:
                series[name] = s
            else:
                try:
                    series[name] = pd.concat([series[name], s])
                except Exception as e:
                    logger.warning(
                        "Not able to concatenate series from the file '%s': %s",
                        file_location.path,
                        e,
                    )
        for s in series.values():
            filtered = s[(s.index >= train_start_date) & (s.index < train_end_date)]
            filtered.sort_index(inplace=True)
            yield filtered



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['gordo_dataset',
 'gordo_dataset.data_provider',
 'gordo_dataset.file_system',
 'gordo_dataset.resources']

package_data = \
{'': ['*']}

install_requires = \
['Inject==4.2.0',
 'PyYAML>=5.3.1,<6.0.0',
 'azure-datalake-store>=0.0.51,<0.0.52',
 'azure-identity>=1.4.0,<2.0.0',
 'azure-storage-file-datalake>=12.1.2,<13.0.0',
 'cachetools>=4.1.0,<5.0.0',
 'cryptography>=3.4.0,<4.0.0',
 'influxdb>=5.3.0,<6.0.0',
 'marshmallow>=3.3.0,<4.0.0',
 'numexpr>=2.7.1,<3.0.0',
 'numpy==1.19.5',
 'pandas>=1.0.3,<2.0.0',
 'pyarrow>=0.17.1,<0.18.0',
 'pydantic>=1.8.2,<2.0.0',
 'scikit-learn>=0.23.2,<0.24.0',
 'typing_extensions>=3.7.4,<4.0.0',
 'xarray>=0.16.2,<0.17.0']

setup_kwargs = {
    'name': 'gordo-dataset',
    'version': '3.0.0.dev2',
    'description': 'Gordo datasets and data providers',
    'long_description': '# Gordo dataset\n\ngordo dataset library essential to build datasets and data providers for [gordo](https://github.com/equinor/gordo) projects.\n\n## Usage\n\n### Data provider\n\nExtend [GordoBaseDataProvider](gordo_dataset/data_provider/base.py) to adapt it to your data source.\n\nSee examples [NcsReader](gordo_data_set/data_provider/ncs_reader.py) that reads either parquet or csv files from Azure Datalake v1.\n\n### Dataset\n\nExtend [GordoBaseDataset](gordo_dataset/base.py).\n\nSee example for [TimeSeriesDataset](gordo_dataset/datasets.py) that arranges the data into consecutive times series.\n\n### Install\n\n`pip install gordo-dataset`\n\n### Uninstall\n\n`pip uninstall gordo-datset`\n',
    'author': 'Equinor ASA',
    'author_email': 'fg_gpl@equinor.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/equinor/gordo-dataset',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



# -*- coding: utf-8 -*-
import collections
import copy
import re
import datetime

import pandas as pd
import dateutil.parser
import logging

from gordo_dataset.sensor_tag import SensorTag


logger = logging.getLogger(__name__)


class BaseDescriptor:
    """
    Base descriptor class

    New object should override __set__(self, instance, value) method to check
    if 'value' meets required needs.
    """

    def __get__(self, instance, owner):
        return instance.__dict__[self.name]

    def __set_name__(self, owner, name):
        self.name = name

    def __set__(self, instance, value):
        raise NotImplementedError("Setting value not implemented for this Validator!")


class ValidDataset(BaseDescriptor):
    """
    Descriptor for attributes requiring type :class:`gordo.workflow.config_elements.Dataset`
    """

    def __set__(self, instance, value):

        # Avoid circular dependency imports
        from gordo_dataset.base import GordoBaseDataset

        if not isinstance(value, GordoBaseDataset):
            raise TypeError(
                f"Expected value to be an instance of GordoBaseDataset, found {value}"
            )
        instance.__dict__[self.name] = value


class ValidDatasetKwargs(BaseDescriptor):
    """
    Descriptor for attributes requiring type :class:`gordo.workflow.config_elements.Dataset`
    """

    def _verify_resolution(self, resolution: str):
        """
        Verifies that a resolution string is supported in pandas
        """
        try:
            pd.tseries.frequencies.to_offset(resolution)
        except ValueError:
            raise ValueError(
                'Values for "resolution" must match pandas frequency terms: '
                "http://pandas.pydata.org/pandas-docs/stable/user_guide/timeseries.html"
            )

    def __set__(self, instance, value):
        if not isinstance(value, dict):
            raise TypeError(f"Expected kwargs to be an instance of dict, found {value}")

        # Check that if 'resolution' is defined, it's one of supported pandas resampling frequencies
        if "resolution" in value:
            self._verify_resolution(value["resolution"])
        instance.__dict__[self.name] = value


class ValidDataProvider(BaseDescriptor):
    """
    Descriptor for DataProvider
    """

    def __set__(self, instance, value):

        # Avoid circular dependency imports
        from gordo_dataset.data_provider.base import GordoBaseDataProvider

        if not isinstance(value, GordoBaseDataProvider):
            raise TypeError(
                f"Expected value to be an instance of GordoBaseDataProvider, "
                f"found {value} "
            )
        instance.__dict__[self.name] = value


class ValidDatetime(BaseDescriptor):
    """
    Descriptor for attributes requiring valid datetime.datetime attribute
    """

    def __set__(self, instance, value):
        datetime_value = None
        if isinstance(value, datetime.datetime):
            datetime_value = value
        elif isinstance(value, str):
            datetime_value = dateutil.parser.isoparse(value)
        else:
            raise ValueError(
                f"'{value}' is not a valid datetime.datetime object or string!"
            )

        if datetime_value.tzinfo is None:
            raise ValueError(f"Provide timezone to timestamp '{value}'")

        instance.__dict__[self.name] = datetime_value


class ValidTagList(BaseDescriptor):
    """
    Descriptor for attributes requiring a non-empty list of strings
    """

    def __set__(self, instance, value):
        if (
            len(value) == 0
            or not isinstance(value, list)
            or not any(isinstance(value[0], inst) for inst in (str, dict, SensorTag))
        ):
            raise ValueError("Requires setting a non-empty list of strings")
        instance.__dict__[self.name] = value



# -*- coding: utf-8 -*-
import logging
from concurrent.futures import ThreadPoolExecutor
from datetime import datetime
import timeit
from typing import Iterable, List, Optional, Tuple, cast, Union

import pandas as pd

from gordo_dataset.file_system.base import FileSystem
from gordo_dataset.sensor_tag import SensorTag, Tag
from gordo_dataset.utils import capture_args
from gordo_dataset.assets_config import AssetsConfig

from .base import GordoBaseDataProvider
from .ncs_contants import NCS_READER_NAME
from .ncs_file_type import load_ncs_file_types, DEFAULT_TYPE_NAMES
from .ncs_lookup import NcsLookup, TagLocations
from .constants import DEFAULT_MAX_FILE_SIZE
from .partition import PartitionBy, split_by_partitions, Partition

from ..exceptions import ConfigException

logger = logging.getLogger(__name__)


class NcsReader(GordoBaseDataProvider):
    @capture_args
    def __init__(
        self,
        storage: FileSystem,
        assets_config: AssetsConfig,
        threads: Optional[int] = 1,
        remove_status_codes: Optional[list] = [0, 64, 60, 8, 24, 3, 32768],
        dl_base_path: Optional[str] = None,
        lookup_for: Optional[List[str]] = None,
        storage_name: Optional[str] = None,
        ncs_lookup: Optional[NcsLookup] = None,
        max_file_size: Optional[int] = DEFAULT_MAX_FILE_SIZE,
        partition_by: Union[str, PartitionBy] = PartitionBy.MONTH,
        **kwargs,  # Do not remove this
    ):
        """
        Creates a reader for tags from the Norwegian Continental Shelf. Currently
        only supports a small subset of assets.

        Parameters
        ----------
        storage: FileSystem
            Storage file system
        assets_config: AssetsConfig
            Assets config
        threads : Optional[int]
            Number of threads to use. If None then use 1 thread
        remove_status_codes: Optional[list]
            Removes data with Status code(s) in the list. By default it removes data
            with Status code 0.
        dl_base_path: Optional[str]
            Base bath used to override the asset to path dictionary. Useful for demos
            and other non-production settings.
        lookup_for:  Optional[List[str]]
            List of file finders by the file type name. Value by default: ``['parquet', 'yearly_parquet', 'csv']``
        storage_name: Optional[str]
            Used by ``AssetsConfig``
        ncs_lookup: Optional[NcsLookup]
            Creates with current ``storage``, ``storage_name`` and ``lookup_for`` if None
        max_file_size: Optional[int]
            Maximal file size
        partition_by: Union[str, PartitionBy]
            Partition by year or month. Default: "month"

        Notes
        -----
        `lookup_for` provide list sorted by priority. It means that for value ``['csv', 'parquet']``
        the reader will prefer to find CSV files over Parquet

        """
        self.storage = storage
        self.assets_config = assets_config

        self.threads = threads
        self.remove_status_codes = remove_status_codes
        self.dl_base_path = dl_base_path

        if lookup_for is None:
            lookup_for = DEFAULT_TYPE_NAMES
        self.lookup_for = lookup_for
        if storage_name is None:
            storage_name = storage.name
        self.storage_name: str = storage_name
        if ncs_lookup is None:
            ncs_lookup = self.create_ncs_lookup(max_file_size)
        elif isinstance(ncs_lookup, NcsLookup):
            raise ConfigException("ncs_lookup should be instance of NcsLookup")
        self.ncs_lookup = ncs_lookup
        self.partition_by = self.prepare_partition_by(partition_by)
        logger.info(f"Starting NCS reader with {self.threads} threads")

    @staticmethod
    def prepare_partition_by(partition_by: Union[str, PartitionBy]) -> PartitionBy:
        result: Optional[PartitionBy]
        if isinstance(partition_by, str):
            result = PartitionBy.find_by_name(partition_by)
            if result is None:
                raise ConfigException("Wrong partition_by argument '%s'" % partition_by)
        else:
            result = partition_by
        return cast(PartitionBy, result)

    def create_ncs_lookup(self, max_file_size: Optional[int]) -> NcsLookup:
        ncs_file_types = load_ncs_file_types(self.lookup_for)
        return NcsLookup(
            self.storage, ncs_file_types, self.storage_name, max_file_size=max_file_size
        )

    @property
    def reader_name(self) -> str:
        """
        Property used for validating result of `AssetsConfig.get_path()`
        """
        return NCS_READER_NAME

    def can_handle_tag(self, tag: Union[str, SensorTag]):
        """
        Implements GordoBaseDataProvider, see base class for documentation
        """
        # TODO do something better here
        return True

    @staticmethod
    def filter_series(
        series: pd.Series,
        train_start_date: datetime,
        train_end_date: datetime,
    ) -> Iterable[pd.Series]:
        return series[
            (series.index >= train_start_date) & (series.index < train_end_date)
        ]

    def load_series(
        self,
        train_start_date: datetime,
        train_end_date: datetime,
        tag_list: List[Union[str, SensorTag]],
        dry_run: Optional[bool] = False,
        **kwargs,
    ) -> Iterable[Tuple[pd.Series, Tag]]:
        """
        See GordoBaseDataProvider for documentation
        """
        if train_end_date < train_start_date:
            raise ValueError(
                f"NCS reader called with train_end_date: {train_end_date} before train_start_date: {train_start_date}"
            )

        partitions = list(
            split_by_partitions(self.partition_by, train_start_date, train_end_date)
        )

        tag_lookups = self.ncs_lookup.assets_config_tags_lookup(
            self.assets_config, tag_list, self.dl_base_path
        )

        if self.threads:
            with ThreadPoolExecutor(max_workers=self.threads) as executor:
                fetched_tags = executor.map(
                    lambda arg: self._load_series_mapper(
                        arg[0], arg[1], partitions, dry_run
                    ),
                    tag_lookups,
                )
                for series, tag in fetched_tags:
                    yield self.filter_series(
                        series, train_start_date, train_end_date
                    ), tag
        else:
            for tag_name, path_with_tag in tag_lookups:
                series, tag = self._load_series_mapper(
                    tag_name, path_with_tag, partitions, dry_run
                )
                yield self.filter_series(series, train_start_date, train_end_date), tag

    def _load_series_mapper(
        self,
        tag_name: str,
        path_with_tag: Optional[Tuple[str, SensorTag]],
        partitions: List[Partition],
        dry_run: Optional[bool] = False,
    ) -> Tuple[pd.Series, Tag]:
        if path_with_tag is None:
            logger.info(
                "Unable to find tag '%s' (asset '%s') directory in storage '%s'",
                tag_name,
                self.storage_name,
            )
            return pd.Series(), tag_name
        tag_dir, tag = path_with_tag
        tag_locations = self.ncs_lookup.files_lookup(tag_dir, tag, partitions)
        return self.read_tag_locations(tag_locations, dry_run), tag

    def read_tag_locations(
        self, tag_locations: TagLocations, dry_run: Optional[bool] = False
    ) -> pd.Series:
        """
        Reads all data from files in ``tag_locations``

        Parameters
        ----------
        tag_locations: TagLocations
        dry_run: bool

        Returns
        -------
        pd.Series

        """
        tag = tag_locations.tag
        partitions = tag_locations.partitions()

        all_partitions = []
        logger.info(f"Downloading tag: {tag} for partitions: {partitions}")
        for tag, partition, location in tag_locations:
            file_path = location.path
            file_type = location.file_type
            logger.info(f"Parsing file {file_path} from partition {partition}")

            try:
                info = self.storage.info(file_path)
                file_size = info.size / (1024 ** 2)
                logger.debug(f"File size for file {file_path}: {file_size:.2f}MB")

                if dry_run:
                    logger.info("Dry run only, returning empty frame early")
                    return pd.Series()
                before_downloading = timeit.default_timer()
                with self.storage.open(file_path, "rb") as f:
                    df = file_type.read_df(f)
                    df = df.rename(columns={"Value": tag.name})
                    df = df[~df["Status"].isin(self.remove_status_codes)]
                    df.sort_index(inplace=True)
                    all_partitions.append(df)
                    logger.info(
                        f"Done in {(timeit.default_timer()-before_downloading):.2f} sec {file_path}"
                    )

            except FileNotFoundError as e:
                logger.debug(f"{file_path} not found, skipping it: {e}")

        try:
            combined = pd.concat(all_partitions)
        except Exception as e:
            logger.debug(f"Not able to concatinate all partitions: {e}.")
            return pd.Series(name=tag.name, data=[])

            # There often comes duplicated timestamps, keep the last
        if combined.index.duplicated().any():
            combined = combined[~combined.index.duplicated(keep="last")]

        return combined[tag.name]

    @staticmethod
    def _verify_tag_path_exist(fs: FileSystem, path: str):
        """
        Verify that the tag path exists, if not the `fs.info` will raise a FileNotFound error.

        Parameters
        ----------
        fs: FileSystem
            File system
        path : str
            Path of tag to be checked if exists.
        """
        fs.info(path)



import os

from typing import Optional
from abc import ABCMeta, abstractmethod
from collections import defaultdict

import gordo_dataset.file_system.adl1 as adl1
import gordo_dataset.file_system.adl2 as adl2

from gordo_dataset.file_system import FileSystem
from gordo_dataset.file_system.adl2 import ADLGen2FileSystem
from gordo_dataset.exceptions import ConfigException
from gordo_dataset.data_provider.secrets_loaders import (
    ADLSecretsLoader,
    ADLEnvSecretsLoader,
)


DEFAULT_STORAGE_TYPE = "adl2"

DEFAULT_SECRETS_LOADER: ADLSecretsLoader = (
    ADLEnvSecretsLoader()
    .from_env("adl1", "dataplatformdlsprod", "DL_SERVICE_AUTH_STR")
    .from_env("adl2", "omniadlseun", "DL2_SERVICE_AUTH_STR")
)


def create_storage(
    storage_type: Optional[str] = None,
    secrets_loader: Optional[ADLSecretsLoader] = None,
    **kwargs
) -> FileSystem:
    """
    Create ``FileSystem`` instance from the config

    Parameters
    ----------
    storage_type: Optional[str]
        Storage type only supported `adl1`, `adl2` values
    secrets_loader: Optional[ADLSecretsLoader]

    kwargs

    Returns
    -------

    """
    if storage_type is None:
        storage_type = DEFAULT_STORAGE_TYPE
    if secrets_loader is None:
        secrets_loader = DEFAULT_SECRETS_LOADER
    if not isinstance(secrets_loader, ADLSecretsLoader):
        raise ConfigException(
            "secrets_loader should be instance of ADLSecretsLoader and not %s type",
            type(secrets_loader),
        )
    storage: FileSystem
    if storage_type == "adl1":
        if "store_name" not in kwargs:
            kwargs["store_name"] = "dataplatformdlsprod"
        kwargs["adl_secret"] = secrets_loader.get_secret(
            storage_type, kwargs["store_name"]
        )
        storage = adl1.ADLGen1FileSystem.create_from_env(**kwargs)
    elif storage_type == "adl2":
        if "account_name" not in kwargs:
            kwargs["account_name"] = "omniadlseun"
        if "file_system_name" not in kwargs:
            kwargs["file_system_name"] = "dls"
        kwargs["adl_secret"] = secrets_loader.get_secret(
            storage_type, kwargs["account_name"]
        )
        storage = adl2.ADLGen2FileSystem.create_from_env(**kwargs)
    else:
        raise ConfigException("Unknown storage type '%s'" % storage_type)
    return storage



import inject

from gordo_dataset.assets_config import AssetsConfig
from gordo_dataset.resource_assets_config import load_assets_config


def config(binder: inject.Binder):
    binder.bind(AssetsConfig, load_assets_config())


def configure_once():
    inject.configure_once(config)



# -*- coding: utf-8 -*-

import abc
import importlib

from datetime import datetime
from copy import copy
from typing import Iterable, List, Optional, Union, Tuple

import pandas as pd

from gordo_dataset.sensor_tag import SensorTag, Tag
from gordo_dataset.exceptions import ConfigException


class GordoBaseDataProvider(object):
    @abc.abstractmethod
    def load_series(
        self,
        train_start_date: datetime,
        train_end_date: datetime,
        tag_list: List[Union[str, SensorTag]],
        dry_run: Optional[bool] = False,
        **kwargs,
    ) -> Iterable[Tuple[pd.Series, Tag]]:
        """
        Load the required data as an iterable of series where each
        contains the values of the tag with time index

        Parameters
        ----------
        train_start_date: datetime
            Datetime object representing the start of fetching data
        train_end_date: datetime
            Datetime object representing the end of fetching data
        tag_list: List[SensorTag]
            List of tags to fetch, where each will end up being its own dataframe
        dry_run: Optional[bool]
            Set to true to perform a "dry run" of the loading.
            Up to the implementations to determine what that means.
        kwargs: Dict
            With these - additional data might be passed by data_provider.

        Returns
        -------
        Iterable[Tuple[pd.Series, SensorTag]]
        """
        ...

    @abc.abstractmethod
    def can_handle_tag(self, tag: Union[str, SensorTag]):
        """
        Returns true if the dataprovider thinks it can possibly read this tag.
        Typically checks if the asset part of the tag is known to the reader.

        Parameters
        ----------
        tag: SensorTag - Dictionary with a "tag" key and optional "asset"

        Returns
        -------
        bool

        """
        ...

    @abc.abstractmethod
    def to_dict(self):
        """
        Serialize this object into a dict representation, which can be used to
        initialize a new object after popping 'type' from the dict.

        Returns
        -------
        dict
        """
        if not hasattr(self, "_params"):
            raise AttributeError(
                "Failed to lookup init parameters, ensure the "
                "object's __init__ is decorated with 'capture_args'"
            )
        # Update dict with the class
        params = self._params
        params_type = ""
        if hasattr(self, "__module__"):
            # Keep back-compatibility
            if self.__module__ != "gordo_dataset.data_provider.providers":
                params_type = self.__module__ + "."
        params_type += self.__class__.__name__
        params["type"] = params_type
        return params

    @classmethod
    @abc.abstractmethod
    def from_dict(cls, config: dict) -> "GordoBaseDataProvider":
        provider_type = "DataLakeProvider"
        if "type" in config:
            config = copy(config)
            provider_type = config.pop("type")

        module = None
        if "." in provider_type:
            module_name, class_name = provider_type.rsplit(".", 1)

            # TODO validate module_name
            try:
                module = importlib.import_module(module_name)
            except ImportError as e:
                raise ConfigException(
                    f"Unable to import module '{module_name}': {str(e)}"
                )
        else:
            from gordo_dataset.data_provider import providers

            module_name, class_name = "gordo_dataset.data_provider", provider_type
            module = providers

        try:
            Provider = getattr(module, class_name)
        except AttributeError:
            raise ConfigException(
                f"Unable to find data provider '{class_name}' in module {module_name}"
            )

        return Provider(**config)

    def get_metadata(self):
        """
        Get metadata about the current state of the data provider
        """
        return dict()
